/* VMS__CVT_PYLONG.C - 29-JUL-1998 Uwe Zessin
   Convert OpenVMS datatypes quadword + octaword to/from Python 'long int'
*/

#include "Python.h"
#include "longintrepr.h"	/* need MASK, etc. */
#include "mymath.h"

#include <lib$routines.h>       /* LIB$name */

/* @@ the following macro copied from LONGOBJECT.C */
#define ABS(x) ((x) < 0 ? -(x) : (x))

/* ------------------------------------------------------------------------- */

/* @@ the following routine was copied from LONGOBJECT.C */
/* it was renamed from 'long_normalize()' to 'vms__cvt_long_normalize()' */

/* Normalize (remove leading zeros from) a long int object.
   Doesn't attempt to free the storage--in most cases, due to the nature
   of the algorithms used, this could save at most be one word anyway. */

static PyLongObject *
vms__cvt_long_normalize(v)
	register PyLongObject *v;
{
	int j = ABS(v->ob_size);
	register int i = j;
	
	while (i > 0 && v->ob_digit[i-1] == 0)
		--i;
	if (i != j)
		v->ob_size = (v->ob_size < 0) ? -(i) : i;
	return v;
} /* vms__cvt_long_normalize */
/* @@ end of copy from LONGOBJECT.C */

/* ------------------------------------------------------------------------- */

/* create a new long int object from an OpenVMS quadword */
/* the quadword is always assumed to be signed */

PyObject *
vms__cvt_quad2pylong (long q_quadword[2])
{
    /* a quadword fits in at most 10 'digits' */
    PyLongObject *v = _PyLong_New(10); /* in LONGOBJECT.C */

    if (v != NULL)
    {
	unsigned long t0 = q_quadword[0];
	unsigned long t1 = q_quadword[1];

	int i;

	for (i = 0; i < 10; i++)
	{
	    v->ob_digit[i] = 0;
	}

	/* handle sign bit */
	if (q_quadword[1] < 0)
	{
	    v->ob_size = -(v->ob_size);
	    t1 = -q_quadword[1];
	    t0 = -q_quadword[0];
	    /* there is no '-0', turn 80000000,00000000	*/
	    /* into '-80000000,00000000'		*/
	    if ((q_quadword[1] == 0x80000000) &&
		(q_quadword[0] == 0x00000000)   )
	    {
		/* no change */
	    }
	    else
	    {
		if (q_quadword[0] != 0)
		{
		    t1--;
		}
		t1 &= 0x7fffffff;	/* mask off sign bit */
	    }
  	}

	/* bits 14:0 */
	v->ob_digit[0] = (digit) (t0 & MASK);
	t0 >>= SHIFT;
	/* bits 29:15 */
	v->ob_digit[1] = (digit) (t0 & MASK);
	t0 >>= SHIFT;
        /* bits 31:30 remain in t0 */

	/* @@ WARNING! This is 'knowledge' how MASK+SHIFT are defined */
	i = t1 & 0x00001fff;	/* get lower 13 bits from t1 */
	i <<= 2;			/* keep room for lower 2 bits of t0 */
	t0 += i;			/* and put into higher part od t0 */
	t1 >>= 13;			/* remove these bits from t1 */

	/* store merged bits */
        v->ob_digit[2] = (digit) (t0 & MASK);

	/* remaining bits from t1 */
	v->ob_digit[3] = (digit) (t1 & MASK);
	t1 >>= SHIFT;

	/* remaining bits from t1 */
	v->ob_digit[4] = (digit) (t1 & MASK);
	t1 >>= SHIFT;

	v->ob_digit[5] = (digit) (t1 & MASK);
	t1 >>= SHIFT;

	v->ob_digit[6] = (digit) (t1 & MASK);
	t1 >>= SHIFT;

	v->ob_digit[7] = (digit) (t1 & MASK);
	t1 >>= SHIFT;

	v = vms__cvt_long_normalize(v); /* copied from LONGOBJECT.C */
    }
    return (PyObject *)v;
} /* vms__cvt_quad2pylong() */

/* ------------------------------------------------------------------------- */

/* convert a python long int object to an OpenVMS quadword */
/* the quadword is always assumed to be signed */
/* a big positive long int can result in a negative quadword ! */

/* returns 1, if OK */

long
vms__cvt_pylong2quad (PyObject * r_longint, unsigned long * q_quadword)
{
	PyObject *vv = r_longint;
	register PyLongObject *v;
	int l_obsize , sign;
	unsigned long t;
	unsigned long sign_bit;
	
	if (vv == NULL || !PyLong_Check(vv))
	{
	    PyErr_BadInternalCall();
	    return -1;
	}

	v = (PyLongObject *)vv;
	l_obsize = v->ob_size;

	*q_quadword      = 0;	/* pre-initialize quadword */
	*(q_quadword +1) = 0;

	if (l_obsize != 0)
	{
	    /* Python 'long int' is not 0 */

	    sign_bit = 0;
	    sign = 1;

	    /* handle sign bit */
	    if (l_obsize < 0)
	    {
		sign = -1;
		l_obsize = -(l_obsize);
	    }

/* mask
6  6   5      5 4        4        3  3     2    2   1      1 0
32109876|54321098|76543210|98765432|10987654|32109876|54321098|76543210

1234                 12345|67890123|45                 1234567|89012345
    1234|56789012|345                 123456|78901234|5
*/
	    /* bits 14:0 */
	    *q_quadword = v->ob_digit[0];

	    /* bits 29:15 */
	    if (l_obsize > 1) /* 2... */
	    {
		t = v->ob_digit[1];
		t <<= SHIFT;
		*q_quadword |= t;
	    }

	    /* bits 31:30 in first longword of quadword */
	    /*  put remaining bits into 2nd longword at 12:0 of quadword */
	    if (l_obsize > 2) /* 3... */
	    {
		t = (v->ob_digit[2] & 3); /* mask lower 2 bits */
		t <<= 30;
		*q_quadword |= t;
		/* ... */
		q_quadword++;	/* next longword */
		t = (v->ob_digit[2] & 0xfffc); /* remove lower 2 bits */
		t >>= 2;		/* shift these bits */
		*q_quadword = t;
	    }
	    else
	    {
		/* later code assumes that pointer is at 2nd longword */
		q_quadword++;	/* next longword */
	    }

	    /* bits 27:12 in second longword of quadword */
	    if (l_obsize > 3) /* 4... */
	    {
		t = v->ob_digit[3];
		t <<= (SHIFT-2); /* 2 bits of prev digit in 1st longword */
		*q_quadword |= t;
	    }

	    /* bits 31:28 in second longword of quadword */
	    /* these are 3 bits! - check sign/for overflow */
	    if (l_obsize > 4) /* 5... */
	    {
		sign_bit = v->ob_digit[4] & 8;
		t = v->ob_digit[4] & 7;
		t <<= (SHIFT+SHIFT-2);
		*q_quadword |= t;

		if ((v->ob_digit[4] & 0x7ff0) != 0)
		{
		    /* @@ overflow */
		    /*	PyErr_SetString(PyExc_OverflowError,
		    ||		"long int too long to convert quadword");
		    */
		    /* @@ set overflow marker */
		    /* don't abort here */
		}
	    }

          if ((*q_quadword == 0) && (*(q_quadword-1) == 0) && (sign_bit != 0))
	  {
		/* maximum negative value - keep it */
		*q_quadword = 0x80000000; /* patch in 'sign_bit' */
	  }
	  else
	  {
	    if ((sign < 0) || (sign_bit != 0))
	    {
		long q_add1[2];
		long l_arraylength = 2;     /* -> lib$addx() */
                long l_status;

		q_add1[0] = 1;
		q_add1[1] = 0;

		*q_quadword = ~(*q_quadword); /* upper longword */
		q_quadword--;	/* pointer to 1st (=low) longword */
		*q_quadword = ~*q_quadword;	/* lower longword */

		/* ignore condition value SS$_INTOVF */
		(void) lib$addx
			(&q_add1[0]	/* addend-array              */
			,q_quadword	/* augend-array              */
			,q_quadword	/* resultant-array           */
			,&l_arraylength /* [,array-length] longwords */
			);
	    }
	    if (l_obsize > 5) /* 6... */
	    {
		/* @@ check remaining bits for overflow */
		    /* @@ overflow */
		    /*	PyErr_SetString(PyExc_OverflowError,
		    ||		"long int too long to convert quadword");
		    */
		    return 0;
	    }
	  }
	}
	return 1;
} /* vms__cvt_pylong2quad() */

/* ------------------------------------------------------------------------- */

/* EOF: VMS__CVT_PYLONG.C */
