/* VMS_LIB.C -- 30-OCT-1996 Uwe Zessin (ZE.)
   interface to (some) OpenVMS LIB$ routines

   06-SEP-1997 ZE. -- changes for new item-codes
   07-SEP-1997 ZE. -- "
   16-OCT-1997 ZE. -- Rename field 'w_vmsver' to 'w_vmsvermin' and add field
 		'w_vmsvermax' because some data can change from version
 		to version.
   02-NOV-1997 ZE. -- convert functions to use PyArg_ParseTuple() instead of
 		PyArg_Parse().
   07-NOV-1997 ZE. -- add GET_COMMON, PUT_COMMON
   08-NOV-1997 ZE. -- add GET_COMMAND, RUN_PROGRAM
   09-NOV-1997 ZE. -- cleanup + some optimizations
   29-NOV-1997 ZE. -- changes for Python V1.5B1
   18-DEC-1997 ZE. -- add FIND_IMAGE_SYMBOL, FREE_EF, GET_EF
   19-DEC-1997 ZE. -- add RESERVE_EF
   21-DEC-1997 ZE. -- add GET_FOREIGN
   31-DEC-1997 ZE. -- add PUT_OUTPUT, ADD_TIMES, SUB_TIMES, CREATE_DIR
   04-JAN-1998 ZE. -- add ASN_WTH_MAX
   09-JAN-1998 ZE. -- add CVT_VECTIM, FID_TO_NAME
   10-JAN-1998 ZE. -- add SET_LOGICAL w/o item-list
   11-JAN-1998 ZE. -- start SET_LOGICAL item-list support
   12-JAN-1998 ZE. -- continue SET_LOGICAL item-list support
   15-JAN-1998 ZE. -- continue SET_LOGICAL, close all memory leaks that
			have (intentionally) been left open during testing
   09-FEB-1998 ZE. -- add remaining DOC strings
   10-FEB-1998 ZE. -- convert day() and day_time() argument 1 to use 64-bit
			system time for compatibility with other routines
   11-AUG-1998 ZE. -- start conversion of 64-bit 'hex-string' to Py long int
   12-AUG-1998 ZE. -- done conversion of 64-bit 'hex-string' to Py long int
*/

#include "python.h"
#include "vmsdef.h"

#include <jpidef.h>		/* JPI$_name */
#include <libdef.h>		/* LIB$_name */
#include <lib$routines.h>	/* LIB$name */
#include <ots$routines.h>	/* OTS$name */
#include <quidef.h>		/* QUI$_name */
#include <ssdef.h>		/* SS$_name */
#include <syidef.h>		/* SYI$_name */
#include <starlet.h>		/* SYS$name, ... */

#include <descrip.h>
typedef struct dsc$descriptor_s dx;
#define initdx(d,v,l) (d)->dsc$w_length = l; \
	(d)->dsc$b_dtype = DSC$K_DTYPE_T; \
	(d)->dsc$b_class = DSC$K_CLASS_S; \
	(d)->dsc$a_pointer = v

/* ------------------------------------------------------------------------- */
extern vmsdef_xw_vmsver vmsdef_gw_vmsver;
/* ------------------------------------------------------------------------- */

static PyObject *vms_lib_Error;		/* exception vms_lib.error */

static char vms_lib_doc__ [] =
"This is an interface to some OpenVMS LIB$ routines.\n\
See the VMS_LIB documentation for details.";

/* table-type-indicator for LIB$GET_SYMBOL / LIB$SET_SYMBOL */
static char* at_sym_global = "GLOBAL";
static char* at_sym_local  = "LOCAL";
#define LIB$K_CLI_LOCAL_SYM	1		/* Local symbol table  */
#define LIB$K_CLI_GLOBAL_SYM	2		/* Global symbol table */

static char* at_lnt_process = "LNM$PROCESS";	/* process logical name table */

static char* at_english = "ENGLISH";		/* english language */

#define S_COMMON_AREA       252
#define	S_DATE_TIME_STRING   23
#define S_SYMBOL_SIZE_MAX  2048

/* ------------------------------------------------------------------------- */
/* function prototypes for external routines */

/* ------------------------------ */

/* input a binary quadword, output a hex string */
/* you MUST provide storage for 16 characters at *t_quadword_hex ! */
long vms__cvt_quad2hex (long* q_quadword_bin, char* t_quadword_hex);

/* ------------------------------ */

/* input a hex string, output a binary quadword */
/* you MUST provide 16 hex characters at *t_quadword_hex ! */
long vms__cvt_hex2quad (char* t_quadword_hex, long* q_quadword_bin);

/* ------------------------------ */

/* input a binary quadword - output a Python long int */
PyObject * vms__cvt_quad2pylong (long q_quadword[2]);

/* ------------------------------ */

/* input a Python long int - output a binary quadword */
long vms__cvt_pylong2quad (PyObject * r_longint, long * q_quadword);

/* ------------------------------------------------------------------------- */
static char vms_lib_add_times__doc[] =
"resultant_time = vms_lib.add_times (time1, time2)\n\
Add Two Quadword Times.";

PyObject *
vms_lib_add_times (PyObject *self, PyObject *args)
{
	PyObject	* ar_time1;
	long		  q_time1[2];

	PyObject	* ar_time2;
	long		  q_time2[2];

	long		  q_resultant_time[2];

	long		  l_status;

	if (!PyArg_ParseTuple(args, "OO", &ar_time1, &ar_time2))
	{
	    return NULL;
	}

	/* ---------- */
	if (!PyLong_Check(ar_time1))
	{
	    PyErr_SetString(PyExc_ValueError, "argument 1: must be long int");
	    return NULL;
	}

	/* convert PyLongInt to binary quadword */
	l_status = vms__cvt_pylong2quad (ar_time1, &q_time1[0]);
	if (l_status == -1)
	{
	    /* error string has been set */
	    return NULL;
	}
	
	/* ---------- */
	if (!PyLong_Check(ar_time2))
	{
	    PyErr_SetString(PyExc_ValueError, "argument 2: must be long int");
	    return NULL;
	}

	/* convert Py long int to binary quadword */
	l_status = vms__cvt_pylong2quad (ar_time2, &q_time2[0]);
	if (l_status == -1)
	{
	    /* error string has been set */
	    return NULL;
	}

	/* ---------- */
	l_status = lib$add_times
	    (&q_time1[0], &q_time2[0], &q_resultant_time[0]);
	if (l_status == LIB$_NORMAL)	/* note: _not_ SS$_NORMAL !! */
	{
	    /* convert binary quadword to py long int */
	    return vms__cvt_quad2pylong (&q_resultant_time[0]);
	}

	/* ---------- */

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_add_times () */

/* ------------------------------------------------------------------------- */

static char vms_lib_asn_wth_mbx__doc[] =
"device-channel,mailbox-channel = \
vms_lib.asn_wth_mbx(device-name,[maximum-message-size],[,buffer-quota])\n\
Assign channel with mailbox.";

PyObject *
vms_lib_asn_wth_mbx (PyObject *self, PyObject *args)
{
	char			* at_device_name;
	struct dsc$descriptor_s   r_device_name;
	long			  l_device_name_len;

	PyObject		* ar_maximum_message_size;
	long			  l_maximum_message_size;
	long			* al_maximum_message_size;

	PyObject		* ar_buffer_quota;
	long			  l_buffer_quota;
	long			* al_buffer_quota;

	unsigned short int	  w_device_channel;
	unsigned short int	  w_mailbox_channel;

	long			  l_status;

	/* ---------- */
	ar_maximum_message_size = NULL;
	ar_buffer_quota		= NULL;
	if (!PyArg_ParseTuple(args, "s#|OO",
	    &at_device_name, &l_device_name_len,
	    &ar_maximum_message_size, &ar_buffer_quota))
	{
	    return NULL;
	}

	/* argument 1: device-name */
	/* string length in a VMS descriptor is limited to 65535 chars */
	if (l_device_name_len > 65535)
	{
	    l_device_name_len = 65535;
	}
	/* set up string descriptor for LIB$ASN_WTH_MBX */
	r_device_name.dsc$w_length  = l_device_name_len;
	r_device_name.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_device_name.dsc$b_class   = DSC$K_CLASS_S;
	r_device_name.dsc$a_pointer = at_device_name;

	/* argument 2: maximum-message-size (optional) */
	if ((ar_maximum_message_size == NULL)   ||
	    (ar_maximum_message_size == Py_None)  )
	{
	    al_maximum_message_size = 0;	/* skip argument */
	}
	else
	{
	    if (!PyInt_Check(ar_maximum_message_size))
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 2: maximum-message-size must be integer or None");
		return NULL;
	    }
	    else
	    {
		l_maximum_message_size = PyInt_AsLong(ar_maximum_message_size);
		al_maximum_message_size = &l_maximum_message_size;
	    }
	}

	/* argument 3: buffer-quota (optional) */
	if ((ar_buffer_quota == NULL) || (ar_buffer_quota == Py_None))
	{
	    al_buffer_quota = 0;		/* skip argument */
	}
	else
	{
	    if (!PyInt_Check(ar_buffer_quota))
	    {
		PyErr_SetString(PyExc_TypeError,
		"argument 3: buffer-quota must be integer or None");
		return NULL;
	    }
	    else
	    {
		l_buffer_quota = PyInt_AsLong(ar_buffer_quota);
		al_buffer_quota = &l_buffer_quota;
	    }
	}

	l_status = lib$asn_wth_mbx
		(&r_device_name
		,al_maximum_message_size
		,al_buffer_quota
		,&w_device_channel
		,&w_mailbox_channel
		);

	if (l_status == SS$_NORMAL)
	{
	    return Py_BuildValue ("ii",
		(int)w_device_channel, (int)w_mailbox_channel);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_asn_wth_mbx() */

/* ------------------------------------------------------------------------- */

static char vms_lib_attach__doc[] =
"vms_lib.attach(pid)\n\
Attach terminal to process.";

PyObject *
vms_lib_attach (PyObject *self, PyObject *args)
{
	long	l_process_id;
	long	l_status;

	if (!PyArg_ParseTuple(args, "l", &l_process_id))
	{
	    return NULL;
	}

	l_status = lib$attach (&l_process_id);

	if (l_status == SS$_NORMAL)
	{
	    Py_INCREF(Py_None);
	    return Py_None;
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_attach () */

/* ------------------------------------------------------------------------- */
static char vms_lib_create_dir__doc[] =
"LIB$CREATE_DIR()\n\
Create a Directory.";

PyObject *
vms_lib_create_dir (PyObject *self, PyObject *args)
{
	char			* at_devdir_spec;
	struct dsc$descriptor_s   r_devdir_spec;
	long			  l_devdir_spec_len;

	PyObject		* ar_owner_uic;
	long			  l_owner_uic;
	long			* al_owner_uic;

	PyObject		* ar_protection_enable;
	long			  l_protection_enable;
	long			* al_protection_enable;

	PyObject		* ar_protection_value;
	long			  l_protection_value;
	long			* al_protection_value;

	PyObject		* ar_maximum_versions;
	long			  l_maximum_versions;
	long			* al_maximum_versions;

	PyObject		* ar_relative_volume_number;
	long			  l_relative_volume_number;
	long			* al_relative_volume_number;

	long			  l_status;

	ar_owner_uic              = NULL;
	ar_protection_enable      = NULL;
	ar_protection_value       = NULL;
	ar_maximum_versions       = NULL;
	ar_relative_volume_number = NULL;
	if (!PyArg_ParseTuple(args, "s#|OOOOO",
	    &at_devdir_spec, &l_devdir_spec_len, &ar_owner_uic,
	    &ar_protection_enable, &ar_protection_value,
	    &ar_maximum_versions, &ar_relative_volume_number))
	{
	    return NULL;
	}

	/* argument 1: device-directory-spec */
	/* string length in a VMS descriptor is limited to 65535 chars */
	if (l_devdir_spec_len > 65535)
	{
	    l_devdir_spec_len = 65535;
	}
	/* set up string descriptor for LIB$CREATE_DIR */
	r_devdir_spec.dsc$w_length  = l_devdir_spec_len;
	r_devdir_spec.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_devdir_spec.dsc$b_class   = DSC$K_CLASS_S;
	r_devdir_spec.dsc$a_pointer = at_devdir_spec;

	/* argument 2: owner-UIC (optional) */
	if ((ar_owner_uic == NULL) || (ar_owner_uic == Py_None))
	{
	    al_owner_uic = 0;		/* skip argument */
	}
	else
	{
	    if (!PyInt_Check(ar_owner_uic))
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 2: owner-UIC must be integer or None");
		return NULL;
	    }
	    else
	    {
		l_owner_uic = PyInt_AsLong(ar_owner_uic);
		al_owner_uic = &l_owner_uic;
	    }
	}

	/* argument 3: protection-enable (optional) */
	if ((ar_protection_enable == NULL) || (ar_protection_enable == Py_None))
	{
	    al_protection_enable = 0;	/* skip argument */
	}
	else
	{
	    if (!PyInt_Check(ar_protection_enable))
	    {
		PyErr_SetString(PyExc_TypeError,
		"argument 3: protection-enable must be 16bit integer or None");
		return NULL;
	    }
	    else
	    {
		l_protection_enable = PyInt_AsLong(ar_protection_enable);
		if (l_protection_enable >= 65536)
		{
		    PyErr_SetString(PyExc_TypeError,
		    "argument 3: protection-enable must be 16bit integer or None");
		    return NULL;
		}
		al_protection_enable = &l_protection_enable;
	    }
	}

	/* argument 4: protection-value (optional) */
	if ((ar_protection_value == NULL) || (ar_protection_value == Py_None))
	{
	    al_protection_value = 0;	/* skip argument */
	}
	else
	{
	    if (!PyInt_Check(ar_protection_value))
	    {
		PyErr_SetString(PyExc_TypeError,
		"argument 4: protection-value must be 16bit integer or None");
		return NULL;
	    }
	    else
	    {
		l_protection_value = PyInt_AsLong(ar_protection_value);
		if (l_protection_value >= 65536)
		{
		    PyErr_SetString(PyExc_TypeError,
		    "argument 4: protection-value must be 16bit integer or None");
		    return NULL;
		}
		al_protection_value = &l_protection_value;
	    }
	}

	/* argument 5: maximum-versions (optional) */
	if ((ar_maximum_versions == NULL) || (ar_maximum_versions == Py_None))
	{
	    al_maximum_versions = 0;	/* skip argument */
	}
	else
	{
	    if (!PyInt_Check(ar_maximum_versions))
	    {
		PyErr_SetString(PyExc_TypeError,
		"argument 4: maximum-versions must be 16bit integer or None");
		return NULL;
	    }
	    else
	    {
		l_maximum_versions = PyInt_AsLong(ar_maximum_versions);
		if (l_maximum_versions >= 65536)
		{
		    PyErr_SetString(PyExc_TypeError,
		    "argument 4: maximum-versions must be 16bit integer or None");
		    return NULL;
		}
		al_maximum_versions = &l_maximum_versions;
	    }
	}

	/* argument 6: relative-volume-number (optional) */
	if ((ar_relative_volume_number == NULL)   ||
	    (ar_relative_volume_number == Py_None)  )
	{
	    al_relative_volume_number = 0;	/* skip argument */
	}
	else
	{
	    if (!PyInt_Check(ar_relative_volume_number))
	    {
		PyErr_SetString(PyExc_TypeError,
		"argument 4: relative-volume-number must be 16bit integer or None");
		return NULL;
	    }
	    else
	    {
		l_relative_volume_number = PyInt_AsLong(ar_relative_volume_number);
		if (l_relative_volume_number >= 65536)
		{
		    PyErr_SetString(PyExc_TypeError,
		    "argument 4: relative-volume-number must be 16bit integer or None");
		    return NULL;
		}
		al_relative_volume_number = &l_relative_volume_number;
	    }
	}

	l_status = lib$create_dir
		(&r_devdir_spec
		,al_owner_uic
		,al_protection_enable
		,al_protection_value
		,al_maximum_versions
		,al_relative_volume_number
		);

	/* this routine can return SS$_NORMAL + SS$_CREATED as success */
	/* you MUST check the return value yourself - no exception is raised! */

	return Py_BuildValue ("i", l_status);
} /* vms_lib_create_dir () */

/* ------------------------------------------------------------------------- */
static char vms_lib_currency__doc[] =
"currency_string = vms_lib.currency ()\n\
Get System Currency Symbol.";

PyObject *
vms_lib_currency (PyObject *self, PyObject *args)
{
	char			t_currency_string[S_SYMBOL_SIZE_MAX+1];
	struct dsc$descriptor_s r_currency_string;
	unsigned short int	w_resultant_length;
	long			l_status;

	/* no arguments taken */
	if (!PyArg_ParseTuple(args, ""))
	{
	    return NULL;
	}

	/* set up descriptor for LIB$CURRENCY */
	r_currency_string.dsc$w_length  = sizeof(t_currency_string);
	r_currency_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_currency_string.dsc$b_class   = DSC$K_CLASS_S;
	r_currency_string.dsc$a_pointer = &t_currency_string[0];

	l_status = lib$currency (&r_currency_string, &w_resultant_length);

	if (l_status == SS$_NORMAL)
	{
	    return Py_BuildValue
		("s#", t_currency_string, (int)w_resultant_length);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_currency () */

/* ------------------------------------------------------------------------- */

static char vms_lib_cvt_vectim__doc[] =
"resultant_time = vms_lib.cvt_vectim(input_time)\n\
Convert 7-Word Vector to Internal Time.";

PyObject *
vms_lib_cvt_vectim (PyObject *self, PyObject *args)
{
	PyObject		* ar_input_time;

	long			  l_tuple_size;
	long			  l_tuple_index;
	PyObject		* ar_tuple_element;
	long			  l_tuple_element;

	unsigned short int	  w_input_time[7];

	long		  q_resultant_time[2];	    /* binary system time */

	long		  l_status;

	/* argument 1: must be a tuple of 7 integers */
	if (!PyArg_ParseTuple(args, "O", &ar_input_time))
	{
	    return NULL;
	}

	if (!PyTuple_Check(ar_input_time))
	{
	    PyErr_SetString(PyExc_TypeError,
		"argument 1: must be a tuple of 7 integers");
	    return NULL;
	}

	l_tuple_size = PyObject_Length(ar_input_time);
	if (l_tuple_size != 7)
	{
	    PyErr_SetString(PyExc_TypeError,
		"argument 1: must be a tuple of 7 integers");
	    return NULL;
	}

	for (l_tuple_index = 0; l_tuple_index < 7; l_tuple_index++)
	{
	    ar_tuple_element = PyTuple_GetItem(ar_input_time, l_tuple_index);
	    if (ar_tuple_element == NULL)
	    {
		return NULL;
	    }
	    if (!PyInt_Check(ar_tuple_element))
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 1: tuple-element:n is not an integer");
		return NULL;
	    }

	    l_tuple_element = PyInt_AsLong(ar_tuple_element);
	    if (PyErr_Occurred())
	    {
		return NULL;
	    }

	    if (l_tuple_element > 65535)
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 1: tuple-element:n is not a 16-bit integer");
		return NULL;
	    }

	    w_input_time[l_tuple_index] = l_tuple_element;
	}

	l_status = lib$cvt_vectim (&w_input_time[0], &q_resultant_time[0]);

	if (l_status == LIB$_NORMAL)
	{
	    /* convert binary quadword to PyLongInt */
	    return vms__cvt_quad2pylong (&q_resultant_time[0]);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_cvt_vectim () */

/* ------------------------------------------------------------------------- */
static char vms_lib_date_time__doc[] =
"date_time_string = vms_lib.date_time ()\n\
Date and Time Returned as a String.";

PyObject *
vms_lib_date_time (PyObject *self, PyObject *args)
{
	char			t_date_time_string[S_DATE_TIME_STRING+1];
	struct dsc$descriptor_s r_date_time_string;
	long			l_date_time_string_len;
	long			l_status;

	/* no arguments taken */
	if (!PyArg_ParseTuple(args, ""))
	{
	    return NULL;
	}

	/* set up descriptor for LIB$DATE_TIME */
	r_date_time_string.dsc$w_length  = S_DATE_TIME_STRING;
	r_date_time_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_date_time_string.dsc$b_class   = DSC$K_CLASS_S;
	r_date_time_string.dsc$a_pointer = &t_date_time_string[0];

	l_status = lib$date_time (&r_date_time_string);

	if (l_status == SS$_NORMAL)
	{
	    l_date_time_string_len = S_DATE_TIME_STRING;
	    return Py_BuildValue
		("s#", t_date_time_string, (int)l_date_time_string_len);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_date_time () */

/* ------------------------------------------------------------------------- */
static char vms_lib_day__doc[] =
"number_of_days, day_time = vms_lib.day ([user-time])\n\
Day Number Returned as a Longword Integer.";

PyObject *
vms_lib_day (PyObject *self, PyObject *args)
{
	PyObject		* ar_user_time;
	long			  q_user_time[2];

	long			  l_number_of_days;
	long			  l_day_time;

	long			  l_status;

	/* argument 1 can be empty, 'None' or PyLongInt */
	ar_user_time = NULL;
	/* allow PyLongInt or None ---v */
	if (!PyArg_ParseTuple(args, "|O", &ar_user_time))
	{
	    /* argument 1 is not:  1) empty,  2) 'None',  3) other */
	    /* -> return with error */
	    return NULL;
	}

	/* argument is 'None' */
	if ((ar_user_time == NULL) || (ar_user_time == Py_None))
	{
	    /* use current date + time to comply with LIB$DAY */
	    l_status = sys$gettim (&q_user_time);
	    if (l_status != SS$_NORMAL)
	    {
		/* failed - error */
		PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
		return NULL;
	    }
	}
        else
	{
	    if (!PyLong_Check(ar_user_time))
	    {
		PyErr_SetString(PyExc_ValueError,
		    "argument 1: must be long int");
		return NULL;
	    }

	    /* convert PyLongInt to binary quadword */
	    l_status = vms__cvt_pylong2quad (ar_user_time, &q_user_time[0]);
	    if (l_status == -1)
	    {
		/* error string has been set */
		return NULL;
	    }
	}

	l_status = lib$day
			(&l_number_of_days
			,&q_user_time
			,&l_day_time
			);

	if (l_status == SS$_NORMAL)
	{
	    return Py_BuildValue ("ii", l_number_of_days, l_day_time);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_day () */

/* ------------------------------------------------------------------------- */
static char vms_lib_day_of_week__doc[] =
"day_number = vms_lib.day_of_week ([user-time])\n\
Show Numeric Day of Week.";

PyObject *
vms_lib_day_of_week (PyObject *self, PyObject *args)
{
	PyObject		* ar_user_time;
	long			  q_user_time[2];

	long			  l_day_number;

	long			  l_status;

	/* argument 1 can be empty, 'None' or PyLongInt */
	ar_user_time = NULL;
	/* allow PyLongInt or None --v */
	if (!PyArg_ParseTuple(args, "|O", &ar_user_time))
	{
	    /* argument 1 is not:  1) empty,  2) 'None',  3) other */
	    /* -> return with error */
	    return NULL;
	}

	/* argument is 'None' */
	if ((ar_user_time == NULL) || (ar_user_time == Py_None))
	{
	    /* use current date + time to comply with LIB$DAY_OF_WEEK */
	    l_status = sys$gettim (&q_user_time);
	    if (l_status != SS$_NORMAL)
	    {
		/* failed - error */
		PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
		return NULL;
	    }
	}
        else
	{
	    if (!PyLong_Check(ar_user_time))
	    {
		PyErr_SetString(PyExc_ValueError,
		    "argument 1: must be long int");
		return NULL;
	    }

	    /* convert PyLongInt to binary quadword */
	    l_status = vms__cvt_pylong2quad (ar_user_time, &q_user_time[0]);
	    if (l_status == -1)
	    {
		/* error string has been set */
		return NULL;
	    }
	}

	l_status = lib$day_of_week (&q_user_time,&l_day_number);
	if (l_status == SS$_NORMAL)
	{
	    return Py_BuildValue("i", l_day_number);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_day_of_week () */

/* ------------------------------------------------------------------------- */
/* @@ code copied from Donn Cave's port */
/* @@ no DOC string */
PyObject *
VMS_LIB_delete_file (PyObject *self, PyObject *args)
{
	/* Another makeshift stopgap.  (RMS directly would be more ideal.) */
	char *file;
	dx filedx;
	int status;

	if (!PyArg_ParseTuple(args, "s", &file))
	{
	    return NULL;
	}
	initdx(&filedx, file, strlen(file));
	status = lib$delete_file (&filedx);
	if (status == SS$_NORMAL) {
		Py_INCREF(Py_None);
		return Py_None;
	} else {
		PyVMS_ErrSetVal(vms_lib_Error, 1, status);
		return NULL;
	}
} /* VMS_LIB_delete_file () */

/* ------------------------------------------------------------------------- */
static char vms_lib_delete_logical__doc[] =
"vms_lib.delete_logical (logical-name [,table-name])\n\
Delete Logical Name.";

PyObject *
vms_lib_delete_logical (PyObject *self, PyObject *args)
{
	char			* at_logical_name;
	struct dsc$descriptor_s   r_logical_name;
	long			  l_logical_name_len;

	char			* at_table_name;
	struct dsc$descriptor_s   r_table_name;
	long			  l_table_name_len;

	long			  l_status;

	/* provide default value for table-name, which is optional */
	at_table_name = at_lnt_process;		/* LNM$PROCESS */

	/* allow string or None --------v */
	if (!PyArg_ParseTuple(args, "s#|z#",
		&at_logical_name, &l_logical_name_len,
		&at_table_name, &l_table_name_len
		))
	{
	    return NULL;
	}

	/* string length in a VMS descriptor is limited to 65535 chars */
	if (l_logical_name_len > 65535)
	{
	    l_logical_name_len = 65535;
	}
	/* set up string descriptor for LIB$DELETE_LOGICAL */
	r_logical_name.dsc$w_length  = l_logical_name_len;
	r_logical_name.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_logical_name.dsc$b_class   = DSC$K_CLASS_S;
	r_logical_name.dsc$a_pointer = at_logical_name;

	/* argument 2 is optional */
	if (at_table_name == NULL)
	{
	    at_table_name = at_lnt_process;	/* default arg 2 */
	}

	/* string length in a VMS descriptor is limited to 65535 chars */
	if (l_table_name_len > 65535)
	{
	    l_table_name_len = 65535;
	}
	/* set up string descriptor for table name */
	r_table_name.dsc$w_length  = l_table_name_len;
	r_table_name.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_table_name.dsc$b_class   = DSC$K_CLASS_S;
	r_table_name.dsc$a_pointer = at_table_name;

	l_status = lib$delete_logical (&r_logical_name, &r_table_name);

	if (l_status == SS$_NORMAL)
	{
	    Py_INCREF(Py_None);
	    return Py_None;
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_delete_logical () */

/* ------------------------------------------------------------------------- */
static char vms_lib_delete_symbol__doc[] =
"vms_lib.delete_symbol (symbol [,table-type-indicator])\n\
Delete CLI Symbol.";

PyObject *
vms_lib_delete_symbol (PyObject *self, PyObject *args)
{
	char			* at_symbol;
	struct dsc$descriptor_s   r_symbol;
	long			  l_symbol_len;

	char			* at_table_type_indicator;
	long			  l_table_type_indicator;
	long			  l_status;

	/* provide default value for table-type-indicator, which is optional */
	at_table_type_indicator = at_sym_local;

	/* allow string or None --------v */
	if (!PyArg_ParseTuple(args, "s#|z",
		&at_symbol, &l_symbol_len, &at_table_type_indicator))
	{
	    return NULL;
	}

	/* string length in a VMS descriptor is limited to 65535 chars */
	if (l_symbol_len > 65535)
	{
	    l_symbol_len = 65535;
	}
	/* set up string descriptor for LIB$DELETE_SYMBOL */
	r_symbol.dsc$w_length  = l_symbol_len;
	r_symbol.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_symbol.dsc$b_class   = DSC$K_CLASS_S;
	r_symbol.dsc$a_pointer = at_symbol;

	/* argument 2 is optional */
	if (at_table_type_indicator == NULL)
	{
	    l_table_type_indicator = LIB$K_CLI_LOCAL_SYM; /* default arg 2 */
	}
	else
	{
	    /* check table indicator */
	    if (strcmp(at_table_type_indicator, at_sym_global) == 0)
	    {
		l_table_type_indicator = LIB$K_CLI_GLOBAL_SYM;
	    }
	    else
	    {
		if (strcmp(at_table_type_indicator, at_sym_local) == 0)
		{
		    l_table_type_indicator = LIB$K_CLI_LOCAL_SYM;
		}
		else
		{
		    /* invalid table type indicator */
		    l_status = LIB$_INVARG;
		    PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
		    return NULL;
		}
	    }
	}

	l_status = lib$delete_symbol (&r_symbol, &l_table_type_indicator);

	if (l_status == SS$_NORMAL)
	{
	    Py_INCREF(Py_None);
	    return Py_None;
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_delete_symbol () */

/* ------------------------------------------------------------------------- */
static char vms_lib_digit_sep__doc[] =
"digit_separator = vms_lib.digit_sep ()\n\
Get Digit Separator Symbol.";

PyObject *
vms_lib_digit_sep (PyObject *self, PyObject *args)
{
	char			t_digit_separator[2049];
	struct dsc$descriptor_s r_digit_separator;
	unsigned short int	w_resultant_length;
	long			l_status;

	/* no arguments taken */
	if (!PyArg_ParseTuple(args, ""))
	{
	    return NULL;
	}

	/* set up string descriptor for LIB$DIGIT_SEP */
	r_digit_separator.dsc$w_length  = sizeof(t_digit_separator);
	r_digit_separator.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_digit_separator.dsc$b_class   = DSC$K_CLASS_S;
	r_digit_separator.dsc$a_pointer = &t_digit_separator[0];

	l_status = lib$digit_sep (&r_digit_separator, &w_resultant_length);
	if (l_status == SS$_NORMAL)
	{
	    return Py_BuildValue
		("s#", t_digit_separator, (int)w_resultant_length);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_digit_sep () */

/* ------------------------------------------------------------------------- */
static char vms_lib_do_command__doc[] =
"vms_lib.do_command (command-string)\n\
Execute Command.";

PyObject *
vms_lib_do_command (PyObject *self, PyObject *args)
{
	char			* at_command_string;
	struct dsc$descriptor_s   r_command_string;
	long			  l_command_string_len;
	long			  l_status;

	if (!PyArg_ParseTuple(args, "s#",
	     &at_command_string, &l_command_string_len))
	{
	    return NULL;
	}

	/* string length in a VMS descriptor is limited to 65535 chars */
	if (l_command_string_len > 65535)
	{
	    l_command_string_len = 65535;
	}
	/* set up string descriptor for LIB$DO_COMMAND */
	r_command_string.dsc$w_length  = l_command_string_len;
	r_command_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_command_string.dsc$b_class   = DSC$K_CLASS_S;
	r_command_string.dsc$a_pointer = at_command_string;

	l_status = lib$do_command (&r_command_string);

	/* any return from LIB$DO_COMMAND indicates an error! */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_do_command () */

/* ------------------------------------------------------------------------- */
static char vms_lib_fid_to_name__doc[] =
"status, acp_status, filespec = vms_lib.fid_to_name (\
device_name, file_id [,directory_id])\n\
Convert Device and File ID to File Specification.";

PyObject *
vms_lib_fid_to_name (PyObject *self, PyObject *args)
{
	char			* at_device_name;
	struct dsc$descriptor_s   r_device_name;
	long			  l_device_name_len;

	PyObject		* ar_file_id;
	unsigned short int	  aw_file_id[3];

	long			  l_tuple_size;
	long			  l_tuple_index;
	PyObject		* ar_tuple_element;
	long			  l_tuple_element;

	char			  t_filespec[256];
	struct dsc$descriptor_s   r_filespec;
	unsigned short int	  w_filespec_length;

	PyObject		* ar_directory_id;
	unsigned short int	  aw_directory_id[3];
	unsigned short int	* aaw_directory_id;

	long			  l_acp_status;

	long			  l_status;

	if (!PyArg_ParseTuple(args, "s#O|O",
	    &at_device_name, &l_device_name_len,
	    &ar_file_id, &ar_directory_id))
	{
	    return NULL;
	}

	/* -------------------- */
	/* argument 1: device-name */
	/* string length in a VMS descriptor is limited to 65535 chars */
	if (l_device_name_len > 65535)
	{
	    l_device_name_len = 65535;
	}
	/* set up string descriptor for LIB$FID_TO_NAME */
	r_device_name.dsc$w_length  = l_device_name_len;
	r_device_name.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_device_name.dsc$b_class   = DSC$K_CLASS_S;
	r_device_name.dsc$a_pointer = at_device_name;

	/* -------------------- */
	/* argument 2: file id */
	if (!PyTuple_Check(ar_file_id))
	{
	    PyErr_SetString(PyExc_TypeError,
		"argument 2: must be a tuple of 3 integers");
	    return NULL;
	}

	l_tuple_size = PyObject_Length(ar_file_id);
	if (l_tuple_size != 3)
	{
	    PyErr_SetString(PyExc_TypeError,
		"argument 2: must be a tuple of 3 integers");
	    return NULL;
	}

	for (l_tuple_index = 0; l_tuple_index < 3; l_tuple_index++)
	{
	    ar_tuple_element = PyTuple_GetItem(ar_file_id, l_tuple_index);
	    if (ar_tuple_element == NULL)
	    {
		return NULL;
	    }
	    if (!PyInt_Check(ar_tuple_element))
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 2: tuple-element:n is not an integer");
		return NULL;
	    }

	    l_tuple_element = PyInt_AsLong(ar_tuple_element);
	    if (PyErr_Occurred())
	    {
		return NULL;
	    }

	    if (l_tuple_element > 65535)
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 2: tuple-element:n is not a 16-bit integer");
		return NULL;
	    }

	    aw_file_id[l_tuple_index] = l_tuple_element;
	}

	/* -------------------- */
	/* argument 3: directory id - optional */
	if (ar_directory_id == NULL)
	{
	    aaw_directory_id = 0;		/* skip argument */
	}
	else
	{
	  if (!PyTuple_Check(ar_directory_id))
	  {
	    PyErr_SetString(PyExc_TypeError,
		"argument 3: must be a tuple of 3 integers");
	    return NULL;
	  }

	  l_tuple_size = PyObject_Length(ar_directory_id);
	  if (l_tuple_size != 3)
	  {
	    PyErr_SetString(PyExc_TypeError,
		"argument 3: must be a tuple of 3 integers");
	    return NULL;
	  }

	  for (l_tuple_index = 0; l_tuple_index < 3; l_tuple_index++)
	  {
	    ar_tuple_element = PyTuple_GetItem(ar_directory_id, l_tuple_index);
	    if (ar_tuple_element == NULL)
	    {
		return NULL;
	    }
	    if (!PyInt_Check(ar_tuple_element))
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 2: tuple-element:n is not an integer");
		return NULL;
	    }

	    l_tuple_element = PyInt_AsLong(ar_tuple_element);
	    if (PyErr_Occurred())
	    {
		return NULL;
	    }

	    if (l_tuple_element > 65535)
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 2: tuple-element:n is not a 16-bit integer");
		return NULL;
	    }

	    aw_directory_id[l_tuple_index] = l_tuple_element;
	  }
          aaw_directory_id = &aw_directory_id[0];
	}

	/* -------------------- */
	/* build descriptor for output */
	r_filespec.dsc$w_length  = sizeof(t_filespec);
	r_filespec.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_filespec.dsc$b_class   = DSC$K_CLASS_S;
	r_filespec.dsc$a_pointer = &t_filespec[0];

	w_filespec_length = 0;
	/* -------------------- */
	l_status = lib$fid_to_name
		(&r_device_name
		,&aw_file_id[0]
		,&r_filespec
		,&w_filespec_length
		,aaw_directory_id
		,&l_acp_status
		);

	return Py_BuildValue ("iis#",
		l_status, l_acp_status,
		&t_filespec[0], (int)w_filespec_length);
} /* vms_lib_fid_to_name () */

/* ------------------------------------------------------------------------- */
/* Find Universal Symbol in Shareable Image File

   The Find Universal Symbol in Shareable Image File routine
   reads universal symbols from the shareable image file. This
   routine then dynamically activates a shareable image into the
   P0 address space of a process. */

static char vms_lib_find_image_symbol__doc[] =
"symbol_value, status = \
vms_lib.find_image_symbol (filename, symbol [,image_name])\n\
Find Universal Symbol in Shareable Image File.";

/* ---------------------------------------- */
long vms_lib__fis
	(struct dsc$descriptor_s * ar_filename
	,struct dsc$descriptor_s * ar_symbol
	,long			 * al_symbol_value
	,struct dsc$descriptor_s * ar_image_name
	)
{
	long l_status;

	/* establish a condition handler to trap any signals and */
	/*  turn them into return values */
	(void) lib$establish (lib$sig_to_ret);
	/* no condition value is returned */

	l_status = lib$find_image_symbol
		(ar_filename
		,ar_symbol
		,al_symbol_value
		,ar_image_name);
	return (l_status);
} /* vms_lib__fis () */
/* ---------------------------------------- */

PyObject *
vms_lib_find_image_symbol (PyObject *self, PyObject *args)
{
	char			* at_filename;
	struct dsc$descriptor_s   r_filename;
	long			  l_filename_len;

	char			* at_symbol;
	struct dsc$descriptor_s   r_symbol;
	long			  l_symbol_len;

	char			* at_image_name;
	struct dsc$descriptor_s   r_image_name;
	long			  l_image_name_len;
	struct dsc$descriptor_s * ar_image_name;

	long			  l_symbol_value;
	long			  l_status;

	at_image_name    = NULL;	/* defaults for optional parameter */
	l_image_name_len = 0;
	if (!PyArg_ParseTuple(args, "s#s#|s#",
	     &at_filename,   &l_filename_len,
	     &at_symbol,     &l_symbol_len,
	     &at_image_name, &l_image_name_len))
	{
	    return NULL;
	}

	/* string length in a VMS descriptor is limited to 65535 chars */
	if (l_filename_len > 65535)
	{
	    l_filename_len = 65535;
	}
	if (l_symbol_len > 65535)
	{
	    l_symbol_len = 65535;
	}

	/* set up string descriptors for LIB$FIND_IMAGE_SYMBOL */
	r_filename.dsc$w_length  = l_filename_len;
	r_filename.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_filename.dsc$b_class   = DSC$K_CLASS_S;
	r_filename.dsc$a_pointer = at_filename;

	r_symbol.dsc$w_length    = l_symbol_len;
	r_symbol.dsc$b_dtype     = DSC$K_DTYPE_T;
	r_symbol.dsc$b_class     = DSC$K_CLASS_S;
	r_symbol.dsc$a_pointer   = at_symbol;

	if (at_image_name != NULL)
	{
	    if (l_image_name_len > 65535)
	    {
		l_image_name_len = 65535;
	    }
	    r_image_name.dsc$w_length  = l_image_name_len;
	    r_image_name.dsc$b_dtype   = DSC$K_DTYPE_T;
	    r_image_name.dsc$b_class   = DSC$K_CLASS_S;
	    r_image_name.dsc$a_pointer = at_image_name;
	    ar_image_name = &r_image_name;
	}
	else
	{
	    ar_image_name = 0;	/* skip parameter */
	}

	l_status = vms_lib__fis
		(&r_filename
		,&r_symbol
		,&l_symbol_value
		,ar_image_name);
	/* return symbol-value and status */
	/* no error is signaled to Python - you MUST check the status ! */
	return Py_BuildValue ("ll", l_symbol_value, l_status);
} /* vms_lib_find_image_symbol () */

/* ------------------------------------------------------------------------- */
static char vms_lib_free_ef__doc[] =
"vms_lib.free_ef (event_flag_number)\n\
Free Event Flag.";

PyObject *
vms_lib_free_ef (PyObject *self, PyObject *args)
{
	long	l_efn;
	long	l_status;

	if (!PyArg_ParseTuple(args, "i", &l_efn))
	{
	    return NULL;
	}

	l_status = lib$free_ef (&l_efn);
	if (l_status == SS$_NORMAL)
	{
	    Py_INCREF(Py_None);
	    return Py_None;
	}

	/* error from LIB$FREE_EF */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_free_ef () */

/* ------------------------------------------------------------------------- */
/* reference to translation table for DVI$_name text/code */
extern struct vmsdef_xr_itmtbl VMSDEF_GR_$DVIDEF[];

static char vms_lib_getdvi__doc[] =
"item_value = vms_lib.getdvi (item_name, channel_number [,device_name])\n\
Get Device/Volume Information.";

PyObject *
vms_lib_getdvi (PyObject *self, PyObject *args)
{
	char			* at_item_code;
	long			  l_item_code_len;
	long			  l_item_code;

	long			  l_channel;
	unsigned short int	  w_channel;
	PyObject		* ar_channel;

	char			* at_device_name;
	struct dsc$descriptor_s   r_device_name;
	long			  l_device_name_len;
	PyObject		* ar_device_name;

	long			  l_integer_value;

#define S_RESULTANT_STRING 32000
	char			  t_resultant_string[S_RESULTANT_STRING+1];
	struct dsc$descriptor_s   r_resultant_string;
	unsigned short int	  w_resultant_length;

	long			  l_trn_index; /* into VMSDEF_GR_$DVIDEF[] */
	long			  l_itm_type;
	long			  l_status;

    l_item_code_len   = 0;
    l_channel         = 0;
    l_device_name_len = 0;
    ar_device_name    = Py_None;	/* default, if omitted */

    if (!PyArg_ParseTuple(args, "s#O|O", &at_item_code, &l_item_code_len,
	&ar_channel, &ar_device_name))
    {
	return NULL;
    }
    else
    {
	if (ar_channel == Py_None)
	{
	    l_channel = -2;		/* indicate skip */
	}
	else
	{
	    if (!PyInt_Check(ar_channel))
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 2: channel number must be integer or None");
		return NULL;
	    }

	    /* @@ this one doesn't work properly? - above is a workaround */
	    /* error mesage is: */
	    /*   TypeError: illegal argument type for built-in operation */
	    if (!PyArg_Parse(ar_channel,
		"l;argument 2: channel number must be integer or None",
		&l_channel))
	    {
		return NULL;
	    }
	}

	if (ar_device_name == Py_None)
	{
	    at_device_name    = NULL;
	    l_device_name_len = 0;
	}
	else
	{
	    if (!PyString_Check(ar_device_name))
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 3: device_name must be string or None");
		return NULL;
	    }

	    /* @@ this one doesn't work properly? - above is a workaround */
	    /* error mesage is: */
	    /*   TypeError: illegal argument type for built-in operation */
	    if (!PyArg_Parse(ar_device_name,
		"s#;argument 3: device_name must be string or None",
		&at_device_name, &l_device_name_len))
	    {
		return NULL;
	    }
	}
    }

    /* ---------------------------------------- *

    /* convert item-code string to DVI$_ numerical value */
    l_item_code = 0;
    l_trn_index = 0;
    while (VMSDEF_GR_$DVIDEF[l_trn_index].w_vmsvermin != 0)
    {
	/* check VMS version */
	if ((VMSDEF_GR_$DVIDEF[l_trn_index].w_vmsvermin <= vmsdef_gw_vmsver) &&
	    (VMSDEF_GR_$DVIDEF[l_trn_index].w_vmsvermax >= vmsdef_gw_vmsver)  )
	{
	    if (strcmp(at_item_code,VMSDEF_GR_$DVIDEF[l_trn_index].at_itmnam_py)
		== 0)
	    {
		/* found textual match - use numerical value of item-code */
		l_item_code = (long)VMSDEF_GR_$DVIDEF[l_trn_index].w_itmcod;
		/* data type */
		if (VMSDEF_GR_$DVIDEF[l_trn_index].c_typ_lib == 'S')
		{
		    l_itm_type = K_ITM_STRING;
		}
		else
		{
		    l_itm_type = K_ITM_INTEGER;
		}
		break;
	    }
	} /* if (VMSDEF_GR_$DVIDEF[l_trn_index].w_vmsvermin/max ... */
	l_trn_index++;				/* check next item code */
    }

    if (VMSDEF_GR_$DVIDEF[l_trn_index].w_vmsvermin == 0)
    {
	/* item code not found - abort */
	PyErr_SetString(PyExc_ValueError, "argument 1: unknown DVI$_ item code");
	return NULL;
    }

    /* ---------------------------------------- *

    /* set up string descriptor for resultant-string argument */
    r_resultant_string.dsc$w_length  = S_RESULTANT_STRING;
    r_resultant_string.dsc$b_dtype   = DSC$K_DTYPE_T;
    r_resultant_string.dsc$b_class   = DSC$K_CLASS_S;
    r_resultant_string.dsc$a_pointer = &t_resultant_string[0];

    /* special marker as arguments cannot be empty */
    /* and device-name normally overrides channel  */
    if (ar_device_name == Py_None)
    {
	w_channel = (short)l_channel;
	/* use channel-number instead of device-name */
	l_status = lib$getdvi
		(&l_item_code, &w_channel
		,0
		,&l_integer_value
		,&r_resultant_string, &w_resultant_length
		);
    }
    else
    {
	/* argument 3 (device-name) is optional */
	/* set up string descriptor for device-name */
	r_device_name.dsc$w_length  = l_device_name_len;
	r_device_name.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_device_name.dsc$b_class   = DSC$K_CLASS_S;
	r_device_name.dsc$a_pointer = at_device_name;

	/* use device-name instead of channel-number */
	l_status = lib$getdvi
		(&l_item_code, 0
		,&r_device_name
		,&l_integer_value
		,&r_resultant_string, &w_resultant_length
		);
    }

    /* check status from LIB$GETDVI() */
    if (l_status == SS$_NORMAL)
    {
	/* adjust string size to ease debugging (EXAMINE/ASCID) */
	r_resultant_string.dsc$w_length = w_resultant_length;

	if (l_itm_type == K_ITM_STRING)
	{
	    /* return resultant-string */
	    return Py_BuildValue
		("s#", t_resultant_string, (int)w_resultant_length);
	}
	else
	{
	    /* return resultant-value (integer / boolean) */
	    return Py_BuildValue ("i", l_integer_value);
	}
    }

    /* error from LIB$GETDVI */
    PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
    return NULL;
} /* vms_lib_getdvi () */

/* ------------------------------------------------------------------------- */
/* reference to translation table for JPI$_name text/code */
extern struct vmsdef_xr_itmtbl VMSDEF_GR_$JPIDEF[];

static char vms_lib_getjpi__doc[] =
"ctx_out, item_value = vms_lib.getjpi (item_name, pid_ctx [,process_name])\n\
Get Job/Process Information.";

PyObject *
vms_lib_getjpi (PyObject *self, PyObject *args)
{
	char			* at_item_code;
	long			  l_item_code_len;
	long			  l_item_code;

	long			  l_process_id;
	PyObject		* ar_process_id;

	char			* at_process_name;
	struct dsc$descriptor_s   r_process_name;
	long			  l_process_name_len;
	PyObject		* ar_process_name;

	long			  q_resultant_value[2];

#define S_RESULTANT_STRING 32000
	char			  t_resultant_string[S_RESULTANT_STRING+1];
	struct dsc$descriptor_s   r_resultant_string;
	unsigned short int	  w_resultant_length;

	long			  l_trn_index; /* into VMSDEF_GR_$JPIDEF[] */
	long			  l_itm_type;
	long			  l_status;

    l_item_code_len    = 0;
    l_process_id       = 0;
    l_process_name_len = 0;
    ar_process_name    = Py_None;	/* default, if omitted */

    if (!PyArg_ParseTuple(args, "s#O|O", &at_item_code, &l_item_code_len,
	&ar_process_id, &ar_process_name))
    {
	return NULL;
    }
    else
    {
	if (ar_process_id == Py_None)
	{
	    /* a PID can never be negative and -1 means a wildcard lookup */
	    l_process_id = -2;		/* skip pid argument */
	}
	else
	{
	    if (!PyInt_Check(ar_process_id))
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 2: process_id must be integer or None");
		return NULL;
	    }

	    /* @@ this one doesn't work properly? - above is a workaround */
	    /* error mesage is: */
	    /*   TypeError: illegal argument type for built-in operation */
	    if (!PyArg_Parse(ar_process_id,
		"l;argument 2 - process_id must be integer or None", 
		&l_process_id))
	    {
		return NULL;
	    }
	}

	if (ar_process_name == Py_None)
	{
	    at_process_name    = NULL;
	    l_process_name_len = 0;
	}
	else
	{
	    if (!PyString_Check(ar_process_name))
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 3: process_name must be string or None");
		return NULL;
	    }
	    /* @@ this one doesn't work properly? - above is a workaround */
	    /* error mesage is: */
	    /*   TypeError: illegal argument type for built-in operation */
	    if (!PyArg_Parse(ar_process_name,
		"s#;argument 3: process_name must be string or None",
		&at_process_name, &l_process_name_len))
	    {
		return NULL;
	    }
	}
    }
    /* -------------------- *
    /* convert item-code string to JPI$_ numerical value */
    l_item_code = 0;
    l_trn_index = 0;
    while (VMSDEF_GR_$JPIDEF[l_trn_index].w_vmsvermin != 0)
    {
	/* check VMS version */
	if ((VMSDEF_GR_$JPIDEF[l_trn_index].w_vmsvermin <= vmsdef_gw_vmsver) &&
	    (VMSDEF_GR_$JPIDEF[l_trn_index].w_vmsvermax >= vmsdef_gw_vmsver)  )
	{
	    if (strcmp(at_item_code,VMSDEF_GR_$JPIDEF[l_trn_index].at_itmnam_py)
		== 0)
	    {
		/* found textual match - use numerical value of item-code */
		l_item_code = (long)VMSDEF_GR_$JPIDEF[l_trn_index].w_itmcod;
		/* data type */
		if (VMSDEF_GR_$JPIDEF[l_trn_index].c_typ_lib == 'S')
		{
		    l_itm_type = K_ITM_STRING;
		}
		else
		{
		    l_itm_type = K_ITM_INTEGER;
		}
		break;
	    }
	}
	l_trn_index++;				/* check next item code */
    }

    if (VMSDEF_GR_$JPIDEF[l_trn_index].w_vmsvermin == 0)
    {
	/* item code not found - abort */
	PyErr_SetString(PyExc_ValueError, "argument 1: unknown JPI$_ item code");
	return NULL;
    }

    /* ---------------------------------------- *

    /* set up string descriptor for resultant-string argument */
    r_resultant_string.dsc$w_length  = S_RESULTANT_STRING;
    r_resultant_string.dsc$b_dtype   = DSC$K_DTYPE_T;
    r_resultant_string.dsc$b_class   = DSC$K_CLASS_S;
    r_resultant_string.dsc$a_pointer = &t_resultant_string[0];

    /* special marker as arguments cannot be empty */
    /* and PID normally overrides process name     */
    if (l_process_id != -2)
    {
	/* use process id instead of name */
	l_status = lib$getjpi
		(&l_item_code, &l_process_id, 0
		,&q_resultant_value[0]
		,&r_resultant_string, &w_resultant_length
		);
    }
    else
    {
	/* set up string descriptor for process-name */
	r_process_name.dsc$w_length  = l_process_name_len;
	r_process_name.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_process_name.dsc$b_class   = DSC$K_CLASS_S;
	r_process_name.dsc$a_pointer = at_process_name;

	/* use process name instead of id */
	l_status = lib$getjpi
		(&l_item_code, 0, &r_process_name
		,&q_resultant_value[0]
		,&r_resultant_string, &w_resultant_length
		);
    }

    /* check status from LIB$GETJPI() */
    if (l_status != SS$_NORMAL)
    {
	/* error from LIB$GETJPI */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
    }

    /* adjust string size to ease debugging (EXAMINE/ASCID) */
    r_resultant_string.dsc$w_length = w_resultant_length;

    /* get PID, even if process name was specified */
    if ((l_item_code == JPI$_PID) && (l_process_id == -2))
    {
	l_process_id = q_resultant_value[0];
    }

    /* special handling for PID with wildcard context */
    if (l_item_code == JPI$_PID)
    {
	long l_pid;

	if (l_process_id > 0)	/* real PID, no wildcard context */
	{
	    l_pid = l_process_id;
	    l_status = SS$_NORMAL;	/* no conversion-error, of course */
	}
	else
	{
	    /* convert (hex) string (of wildcard context) to integer */
	    l_status = ots$cvt_tz_l
		(&r_resultant_string	/* fixed-or-dynamic-input-string */
		,&l_pid			/* varying-output-value		 */
		,4			/* [,output-value-size]		 */
		,0			/* [,flags-value]		 */
		);			/*  0=interpret blanks as zeroes */
	}

	if (l_status == SS$_NORMAL)
	{
	    /* return PID/context-value and numerical PID */
	    return Py_BuildValue ("ii", l_process_id, l_pid);
	}
	else
	{
	    /* indicate conversion-error - this should NOT happen !! */
	    PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	    return NULL;
	}
    }
    else
    {
	/* not using wildcard context */
	if (l_itm_type == K_ITM_STRING)
	{
	    /* return PID/context-value and resultant-string */
	    return Py_BuildValue
		("is#", l_process_id,
		t_resultant_string, (int)w_resultant_length);
	}
	else
	{
	    /* return PID/context-value and resultant-value */
	    return Py_BuildValue("ii", l_process_id, q_resultant_value[0]);
	}
    }
} /* vms_lib_getjpi () */

/* ------------------------------------------------------------------------- */

/* reference to translation table for QUI$_name text/code */
extern struct vmsdef_xr_itmtbl VMSDEF_GR_$QUIDEF[];

static char vms_lib_getqui__doc[] =
"result = vms_lib.getqui (function_code, item_name, \
search_number, search_name, search_flags)\n\
Get Queue Information.";

PyObject *
vms_lib_getqui (PyObject *self, PyObject *args)
{
	char			* at_function_code;
	long			  l_function_code_len;
	long			  l_function_code;

	char			* at_item_code;
	long			  l_item_code_len;
	long			  l_item_code;

	long			  l_search_number;
	PyObject		* ar_search_number;

	char			* at_search_name;
	struct dsc$descriptor_s   r_search_name;
	long			  l_search_name_len;

	long			  l_search_flags;

	long			  q_resultant_value[2];

#define S_RESULTANT_STRING 32000
	char			  t_resultant_string[S_RESULTANT_STRING+1];
	struct dsc$descriptor_s   r_resultant_string;
	unsigned short int	  w_resultant_length;

	long			  l_trn_index; /* into VMSDEF_GR_$QUIDEF[] */
	long			  l_itm_type;
	long			  l_status;

	long			  x_onearg;	/* only one argment given */

/* arguments:	function-code		"QUI$_DISPLAY_ENTRY"
		[,item-code]		"QUI$_JOB_NAME"
		[,search-number]	entry-number
		[,search-name]
		[,search-flags]		 QUI$_SEARCH_NUMBER ?
		[,resultant-value]
		[,resultant-string]
		[,resultant-length]

  The following table describes all of the required and optional
  input needed to perform your task with LIB$GETQUI.

Function                     Input Description

QUI$_CANCEL                  Accepts no input.

QUI$_DISPLAY_CHARACTERISTIC  A characteristic name or number, or both.
                             Optionally, a search flags number.

QUI$_DISPLAY_ENTRY           Optionally, an entry number, user name, and search
                             flags number. The default user name is that of the
                             calling process.

QUI$_DISPLAY_FILE            Optionally, a search flags number.

QUI$_DISPLAY_FORM            A form name or number, or both. Optionally, a
                             search flags number.

QUI$_DISPLAY_JOB             Optionally, a search flags number.

QUI$_DISPLAY_QUEUE           A queue name. Optionally, a search flags number.

QUI$_TRANSLATE_QUEUE         A queue name.

*/
    l_function_code_len = 0;
    l_item_code_len     = 0;
    l_search_number     = 0;
    l_search_name_len   = 0;
    l_search_flags      = 0;

    /* this is for QUI$_CANCEL_OPERATION */
    if (PyArg_ParseTuple(args, "s#",
	    &at_function_code, &l_function_code_len))
    {
	x_onearg = 1;
    }
    else
    {
	PyErr_Clear();
	if (!PyArg_ParseTuple(args, "s#s#|ls#l"
		,&at_function_code, &l_function_code_len
		,&at_item_code, &l_item_code_len
		,&l_search_number
		,&at_search_name, &l_search_name_len
		,&l_search_flags
		))
	{
	    PyErr_Clear();
	    if (!PyArg_ParseTuple(args, "s#s#|Os#l"
		,&at_function_code, &l_function_code_len
		,&at_item_code, &l_item_code_len
		,&ar_search_number
		,&at_search_name, &l_search_name_len
		,&l_search_flags
		))
	    {
		return NULL;
	    }
	    else
	    {
		if (ar_search_number == Py_None)
		{
		    l_search_number = -2; /* skip search_number argument */
		}
		else
		{
		    PyErr_SetString(PyExc_TypeError,
			"argument 3: search_number must be number or None");
		    return NULL;
		}
	    }
	}
	x_onearg = 0;
    }

    /* ---------------------------------------- */

    /* convert function-code string to QUI$_ numerical value */
    l_function_code = 0;
    l_trn_index = 0;
    while (VMSDEF_GR_$QUIDEF[l_trn_index].w_vmsvermin != 0)
    {
	/* check VMS version */
	if ((VMSDEF_GR_$QUIDEF[l_trn_index].w_vmsvermin <= vmsdef_gw_vmsver) &&
	    (VMSDEF_GR_$QUIDEF[l_trn_index].w_vmsvermax >= vmsdef_gw_vmsver)  )
	{
	    if (strcmp(at_function_code,VMSDEF_GR_$QUIDEF[l_trn_index].at_itmnam_py)
		== 0)
	    {
		/* check if item is really a function code item 'R' */
		if (! (VMSDEF_GR_$QUIDEF[l_trn_index].l_flags &
			ITMTBL_M_GETQUI_FNCREQ)
		   )
		{
		    /* wrong type of function code - abort */
		    PyErr_SetString(PyExc_ValueError,
			"argument 1: not a QUI$_ request function code");
		    return NULL;
		}
		/* found textual match - use numerical value of function-code */
		l_function_code = (long)VMSDEF_GR_$QUIDEF[l_trn_index].w_itmcod;
		break;
	    }
	}
	l_trn_index++;				/* check next function code */
    }

    /* ---------------------------------------- */
    if (VMSDEF_GR_$QUIDEF[l_trn_index].w_vmsvermin == 0)
    {
	/* function code not found - abort */
	PyErr_SetString(PyExc_ValueError, "argument 1: unknown QUI$_ function code");
	return NULL;
    }
    /* ---------------------------------------- */

    /* CANCEL_OPERATION uses only one argument */
    if (l_function_code == QUI$_CANCEL_OPERATION)
    {
	if (!x_onearg)
	{
	    PyErr_SetString(PyExc_TypeError,
		"QUI$_CANCEL_OPERATION requires exactly one argument");
	    return NULL;
	}
	else
	{
	    /* This crashes the program on some VMS version when	*/
	    /* QUI$_CANCEL_OPERATION is used after a different function */
	    /* l_status = lib$getqui (&l_function_code); */

	    /* this appears to work */
	    l_status = lib$getqui
		(&l_function_code ,0
		,0 ,0
		,0
		,0
		,0 ,0
		);
	    if (l_status == SS$_NORMAL)
	    {
		Py_INCREF(Py_None);
		return Py_None;
	    }
	    else
	    {
		/* error from LIB$GETQUI */
		PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
		return NULL;
	    }
	}
    }

    /* ---------------------------------------- */

    /* all other functions require more arguments */
    if (x_onearg)
    {
	PyErr_SetString(PyExc_TypeError,
	    "function requires at least two arguments");
	return NULL;
    }

    /* ---------------------------------------- */

    /* convert item-code string to QUI$_ numerical value */
    l_item_code = 0;
    l_trn_index = 0;
    while (VMSDEF_GR_$QUIDEF[l_trn_index].w_vmsvermin != 0)
    {
	/* check VMS version */
	if ((VMSDEF_GR_$QUIDEF[l_trn_index].w_vmsvermin <= vmsdef_gw_vmsver) &&
	    (VMSDEF_GR_$QUIDEF[l_trn_index].w_vmsvermax >= vmsdef_gw_vmsver)  )
	{
	    if (strcmp(at_item_code,VMSDEF_GR_$QUIDEF[l_trn_index].at_itmnam_py)
		== 0)
	    {
		/* check that this item is not a request code */
		if ((VMSDEF_GR_$QUIDEF[l_trn_index].l_flags &
			ITMTBL_M_GETQUI_FNCREQ)
		   )
		{
		    /* wrong type of function code - abort */
		    PyErr_SetString(PyExc_ValueError,
			"argument 2: wrong QUI$_ item name for this argument");
		    return NULL;
		}
		/* found textual match - use numerical value of item-code */
		l_item_code = (long)VMSDEF_GR_$QUIDEF[l_trn_index].w_itmcod;
		/* data type */
		if (VMSDEF_GR_$QUIDEF[l_trn_index].c_typ_lib == 'S')
		{
		    l_itm_type = K_ITM_STRING;
		}
		else
		{
		    l_itm_type = K_ITM_INTEGER;
		}
		break;
	    }
	}
	l_trn_index++;				/* check next item code */
    }

    if (VMSDEF_GR_$QUIDEF[l_trn_index].w_vmsvermin == 0)
    {
	/* item-code not found - abort */
	PyErr_SetString(PyExc_ValueError, "argument 2: unknown QUI$_ item code");
	return NULL;
    }

    /* ---------------------------------------- */

    /* set up string descriptor for resultant-string argument */
    r_resultant_string.dsc$w_length  = S_RESULTANT_STRING;
    r_resultant_string.dsc$b_dtype   = DSC$K_DTYPE_T;
    r_resultant_string.dsc$b_class   = DSC$K_CLASS_S;
    r_resultant_string.dsc$a_pointer = &t_resultant_string[0];

    /* special marker as arguments cannot be empty */
    /* and number normally overrides name	   */
    if (l_search_number != -2)
    {
	/* use search-number instead of search-name */
	l_status = lib$getqui
		(&l_function_code ,&l_item_code
		,&l_search_number ,0
		,&l_search_flags
		,&q_resultant_value[0]
		,&r_resultant_string ,&w_resultant_length
		);
    }
    else
    {
	/* set up string descriptor for search-name */
	r_search_name.dsc$w_length  = l_search_name_len;
	r_search_name.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_search_name.dsc$b_class   = DSC$K_CLASS_S;
	r_search_name.dsc$a_pointer = at_search_name;

	/* use search-name instead of search-number */
	l_status = lib$getqui
		(&l_function_code ,&l_item_code
		,0 ,&r_search_name
		,&l_search_flags
		,&q_resultant_value[0]
		,&r_resultant_string ,&w_resultant_length
		);
    }

    /* check status from LIB$GETQUI() */
    if (l_status == SS$_NORMAL)
    {
	/* adjust string size to ease debugging (EXAMINE/ASCID) */
	r_resultant_string.dsc$w_length = w_resultant_length;

	if (l_itm_type == K_ITM_STRING)
	{
	    /* return resultant-string */
	    return Py_BuildValue
		("s#", t_resultant_string, (int)w_resultant_length);
	}
	else
	{
	    /* return resultant-value */
	    return Py_BuildValue ("i", q_resultant_value[0]);
	}
    }

    /* failed - error */
    PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
    return NULL;
} /* vms_lib_getqui () */

/* ------------------------------------------------------------------------- */
/* reference to translation table for SYI$_name text/code */
extern struct vmsdef_xr_itmtbl VMSDEF_GR_$SYIDEF[];

static char vms_lib_getsyi__doc[] =
"ctx_out, item_value = vms_lib.getsyi (item_name, csid_ctx [,node_name])\n\
Get Systemwide Information.";

PyObject *
vms_lib_getsyi (PyObject *self, PyObject *args)
{
	char			* at_item_code;
	long			  l_item_code_len;
	long			  l_item_code;

	long			  l_system_id;
	PyObject		* ar_system_id;

	char			* at_node_name;
	struct dsc$descriptor_s   r_node_name;
	long			  l_node_name_len;

	long			  q_resultant_value[2];

#define S_RESULTANT_STRING 32000
	char			  t_resultant_string[S_RESULTANT_STRING+1];
	struct dsc$descriptor_s   r_resultant_string;
	unsigned short int	  w_resultant_length;

	long			  l_trn_index; /* into VMSDEF_GR_$SYIDEF[] */
	long			  l_itm_type;
	long			  l_status;

/* arguments:	s:item-code (e.g. "NODENAME")
		i:system-cluster-id	[*]
		s:node-name, optional

[*] Note: do not confuse the cluster system id (CSID) with the contents
    of the system identification register (SID)
*/

    l_item_code_len = 0;
    l_system_id = 0;
    l_node_name_len = 0;

    if (!PyArg_ParseTuple(args, "s#l|s#", &at_item_code, &l_item_code_len,
	&l_system_id, &at_node_name, &l_node_name_len))
    {
	PyErr_Clear();
	if (!PyArg_ParseTuple(args, "s#O|s#",
	    &at_item_code, &l_item_code_len,
	    &ar_system_id, &at_node_name, &l_node_name_len))
	{
	    return NULL;
	}
	else
	{
	    if (ar_system_id == Py_None)
	    {
		l_system_id = -2;		/* skip csid argument */
	    }
	    else
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 2: cluster system id must be number or None");
		return NULL;
	    }
	}
    }

    /* convert item-code string to SYI$_ numerical value */
    l_item_code = 0;
    l_trn_index = 0;
    while (VMSDEF_GR_$SYIDEF[l_trn_index].w_vmsvermin != 0)
    {
	/* check VMS version */
	if ((VMSDEF_GR_$SYIDEF[l_trn_index].w_vmsvermin <= vmsdef_gw_vmsver) &&
	    (VMSDEF_GR_$SYIDEF[l_trn_index].w_vmsvermax >= vmsdef_gw_vmsver)  )
	{
	    if (strcmp(at_item_code,VMSDEF_GR_$SYIDEF[l_trn_index].at_itmnam_py)
		== 0)
	    {
		/* found textual match - use numerical value of item-code */
		l_item_code = (long)VMSDEF_GR_$SYIDEF[l_trn_index].w_itmcod;
		/* data type */
		if (VMSDEF_GR_$SYIDEF[l_trn_index].c_typ_lib == 'S')
		{
		    l_itm_type = K_ITM_STRING;
		}
		else
		{
		    l_itm_type = K_ITM_INTEGER;
		}
		break;
	    }
	}
	l_trn_index++;				/* check next item code */
    }

    if (VMSDEF_GR_$SYIDEF[l_trn_index].w_vmsvermin == 0)
    {
	/* item code not found - abort */
	PyErr_SetString(PyExc_ValueError, "argument 1: unknown SYI$_ item code");
	return NULL;
    }

    /* set up string descriptor for resultant-string argument */
    r_resultant_string.dsc$w_length  = S_RESULTANT_STRING;
    r_resultant_string.dsc$b_dtype   = DSC$K_DTYPE_T;
    r_resultant_string.dsc$b_class   = DSC$K_CLASS_S;
    r_resultant_string.dsc$a_pointer = &t_resultant_string[0];

    q_resultant_value[0] = 0;
    q_resultant_value[1] = 0;

    /* special marker as arguments cannot be empty and */
    /* and CSID normally overrides node name           */
    if (l_system_id != -2)
    {
	/* use CSID instead of node name */
	l_status = lib$getsyi
		(&l_item_code, &q_resultant_value[0]
		,&r_resultant_string,&w_resultant_length
		,&l_system_id, 0
		);
    }
    else
    {
	/* set up string descriptor for node-name */
	r_node_name.dsc$w_length  = l_node_name_len;
	r_node_name.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_node_name.dsc$b_class   = DSC$K_CLASS_S;
	r_node_name.dsc$a_pointer = at_node_name;

	/* use node name instead of CSID */
	l_status = lib$getsyi
		(&l_item_code, &q_resultant_value[0]
		,&r_resultant_string, &w_resultant_length
		,0, &r_node_name
		);
    }

    /* check status from LIB$GETSYI() */
    if (l_status == SS$_NORMAL)
    {
	/* adjust string size to ease debugging (EXAMINE/ASCID) */
	r_resultant_string.dsc$w_length = w_resultant_length;

	/* get CSID, even if node name was specified */
	if ((l_item_code == SYI$_NODE_CSID) && (l_system_id == -2))
	{
	    l_system_id = q_resultant_value[0];
	}

	/* remove skip marker */
	/* @@ need to be tested in a cluster ! */
	if (l_system_id == -2)
	{
	    l_system_id = 0;
	}

	/* special handling for CSID with wildcard context */
	if (l_item_code == SYI$_NODE_CSID)
	{
	    long l_csid;

	    if (l_system_id > 0)	/* @@ necessary ? */
	    {
		l_csid = l_system_id;
		l_status = SS$_NORMAL; /* no conversion-error, of course */
	    }
	    else
	    {
		/* convert (hex) string to integer */
		l_status = ots$cvt_tz_l
		    (&r_resultant_string  /* fixed-or-dynamic-input-string */
		    ,&l_csid		  /* varying-output-value	   */
		    ,4			  /* [,output-value-size]	   */
		    ,0			  /* [,flags-value]		   */
		    );			  /*  0=interpret blanks as zeroes */
	    }

	    if (l_status == SS$_NORMAL)
	    {
		/* return CSID/context-value and numerical CSID */
		return Py_BuildValue ("ii", l_system_id, l_csid);
	    }
	    else
	    {
		/* indicate conversion-error - this should NOT happen !! */
		PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
		return NULL;
	    }
	}
	else
	{
	    if (l_itm_type == K_ITM_STRING)
	    {
		/* return CSID/context-value and resultant-string */
		return Py_BuildValue
			("is#", l_system_id,
			t_resultant_string, (int)w_resultant_length);
	    }
	    else
	    {
		/* return CSID/context-value and resultant-value */
		return Py_BuildValue
			("ii", l_system_id, q_resultant_value[0]);
	    }
	}
    }

    /* error from LIB$GETSYI() */
    PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
    return NULL;
} /* vms_lib_getsyi () */

/* ------------------------------------------------------------------------- */
static char vms_lib_get_command__doc[] =
"data_from_command, resultant_length = \
vms_lib.get_command (prompt [,buffer_length])\n\
Get Line from SYS$COMMAND.";

PyObject *
vms_lib_get_command (PyObject *self, PyObject *args)
{
#define S_COMMAND_BUFFER 65535
	char			* at_prompt_string;
	struct dsc$descriptor_s   r_prompt_string;
	long 			  l_prompt_length;
	void			* ar_prompt_string;

	char			* at_buffer;
	PyObject		* ar_buffer;

	long 			  l_command_buffer;
	long 			  l_resultant_length;
	struct dsc$descriptor_s   r_resultant_string;
	unsigned short int 	  w_resultant_length;

	long			  l_status;

	/* argument 1 can be empty, 'None' or (prompt) string */
	at_prompt_string = NULL;
	/* argument 2 is optionally 'int', default=65535 */
	l_command_buffer = S_COMMAND_BUFFER;

	if (!PyArg_ParseTuple(args, "|z#i",
	    &at_prompt_string, &l_prompt_length, &l_command_buffer))
	{
	    return NULL;
	}

	if (at_prompt_string == NULL)
	{
	    ar_prompt_string = 0;	       /* no prompt string requested */
	}
	else
	{
	    /* string length in a VMS descriptor is limited to 65535 chars */
	    if (l_prompt_length > 65535)
	    {
		PyErr_SetString(PyExc_ValueError,
	    "argument 1: prompt string length is limited to 65535 characters");
		return NULL;
	    }
	    /* set up descriptor for LIB$GET_COMMAND */
	    r_prompt_string.dsc$w_length  = l_prompt_length;
	    r_prompt_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	    r_prompt_string.dsc$b_class   = DSC$K_CLASS_S;
	    r_prompt_string.dsc$a_pointer = at_prompt_string;

	    ar_prompt_string = &r_prompt_string;
	}

	if (l_command_buffer > S_COMMAND_BUFFER)
	{
	    /* string length in a VMS descriptor is limited to 65535 chars */
	    PyErr_SetString(PyExc_ValueError,
	     "argument 2: command buffer size is limited to 65535 characters");
	    return NULL;
	}

	/* allocate input buffer */
	at_buffer = malloc (l_command_buffer);
	if (at_buffer == NULL)
	{
	    PyErr_NoMemory();
	    return NULL;
	}

	/* set up descriptor for LIB$GET_COMMAND */
	r_resultant_string.dsc$w_length  = l_command_buffer;
	r_resultant_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_resultant_string.dsc$b_class   = DSC$K_CLASS_S;
	r_resultant_string.dsc$a_pointer = at_buffer;

	l_status = lib$get_command
	    (&r_resultant_string
	    ,ar_prompt_string		/* optional */
	    ,&w_resultant_length
	    );

	/* note: string truncation is not considered an error!  */
	/* Please check 2nd value returned			*/
	if ((l_status == SS$_NORMAL) || (l_status == LIB$_INPSTRTRU))
	{
	    long l_resultant_length;

	    /* create string object for data received from SYS$COMMAND */
	    ar_buffer = PyString_FromStringAndSize
		((char *)NULL, (int)w_resultant_length);
	    if (ar_buffer == NULL)
	    {
		/* release temporary buffer */
		(void)free (at_buffer);

		return NULL;		/* object creation failed */
	    }
	    /* copy data into string object */
	    (void) memcpy
		(PyString_AsString(ar_buffer)
		,at_buffer
		,(size_t)w_resultant_length
		);

	    /* release temporary buffer */
	    (void)free (at_buffer);

	    /* change resultant length to a negative value */
	    /*   if string truncation occured		   */
	    if (l_status == LIB$_INPSTRTRU)
	    {
		l_resultant_length = 0- (int)w_resultant_length;
	    }
	    else
	    {
		l_resultant_length = (int)w_resultant_length;
	    }

	    return Py_BuildValue ("Oi", ar_buffer, l_resultant_length);
	}

	/* release temporary buffer */
	(void)free (at_buffer);

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_get_command () */

/* ------------------------------------------------------------------------- */
static char vms_lib_get_common__doc[] =
"resultant_string, resultant_length = vms_lib.get_common ()\n\
Get String from Common.";

PyObject *
vms_lib_get_common (PyObject *self, PyObject *args)
{
	char			  t_source_string[S_COMMON_AREA]; /* no .ASCIZ */
	struct dsc$descriptor_s   r_source_string;
	unsigned short int 	  w_resultant_length;

	PyObject		* ar_buffer;

	long			  l_status;

	/* no input argument */
	if (!PyArg_ParseTuple(args, ""))
	{
	    return NULL;
	}

	/* set up descriptor for LIB$GET_COMMON */
	r_source_string.dsc$w_length  = S_COMMON_AREA;
	r_source_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_source_string.dsc$b_class   = DSC$K_CLASS_S;
	r_source_string.dsc$a_pointer = &t_source_string[0];

	l_status = lib$get_common (&r_source_string, &w_resultant_length);

	/* Warning - string truncation is considered an error, too */
	/* this indicates a programmer error or a change in VMS */
	if (l_status == SS$_NORMAL)
	{
	    /* create a string object which might include '\0' */
	    ar_buffer = PyString_FromStringAndSize
		((char *)NULL, (int)w_resultant_length);
	    if (ar_buffer == NULL)
	    {
		return NULL;
	    }
	    /* copy data into string object */
	    (void) memcpy
		(PyString_AsString(ar_buffer)
		,&t_source_string[0]
		,(size_t)w_resultant_length
		);
	    return Py_BuildValue ("Oi", ar_buffer, (int)w_resultant_length);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_get_common () */

/* ------------------------------------------------------------------------- */
static char vms_lib_get_ef__doc[] =
"event_flag_number = vms_lib.get_ef ()\n\
Get Event Flag.";

PyObject *
vms_lib_get_ef (PyObject *self, PyObject *args)
{
	long	l_efn;
	long	l_status;

	/* no input argument */
	if (!PyArg_ParseTuple(args, ""))
	{
	    return NULL;
	}

	l_status = lib$get_ef (&l_efn);
	if (l_status == SS$_NORMAL)
	{
	    return Py_BuildValue ("i", l_efn);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_get_ef () */

/* ------------------------------------------------------------------------- */
static char vms_lib_get_foreign__doc[] =
"resultant_string, flags_out = vms_lib.get_foreign ([prompt_string] [,flags])\n\
Get Foreign Command Line.";

PyObject *
vms_lib_get_foreign (PyObject *self, PyObject *args)
{
#ifndef S_RESULTANT_STRING
#define S_RESULTANT_STRING 32000
#endif
	char			  t_resultant_string[S_RESULTANT_STRING+1];
	char			* at_resultant_string;
	struct dsc$descriptor_s   r_resultant_string;
	long			  l_resultant_string_len;

	unsigned short int	  w_resultant_length;

	char			* at_prompt_string;
	struct dsc$descriptor_s   r_prompt_string;
	struct dsc$descriptor_s * ar_prompt_string;
	long			  l_prompt_string_len;

	long			  l_flags;
	long			* al_flags;
	PyObject		* ar_flags;

	long			  l_status;

	/* argument 1 can be empty, 'None' or string */
	if (PyArg_ParseTuple(args, ""))
	{
	  /* no arguments given, set defaults */
	  ar_prompt_string = NULL;
	  al_flags         = 0;				/* default */
	}
	else
	{
	  PyErr_Clear();
	  ar_flags = NULL; /* set defaults for optional argument */
          /* allow string or None -----v  v-- flags is int or none or opt. */
	  if (!PyArg_ParseTuple(args, "z#|O",
		&at_prompt_string, &l_prompt_string_len,
		&ar_flags
		))
	  {
	    return NULL;
	  }

	  /* argument 1 processing */
	  if (at_prompt_string == NULL)	/* argument is 'None' */
	  {
	    ar_prompt_string = 0;	/* skip argument */
	  }
	  else
	  {
	    /* string length in a VMS descriptor is limited to 65535 chars */
	    if (l_prompt_string_len > 65535)
	    {
	      l_prompt_string_len = 65535;
	    }
            /* set up string descriptor for LIB$GET_FOREIGN */
	    r_prompt_string.dsc$w_length  = l_prompt_string_len;
	    r_prompt_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	    r_prompt_string.dsc$b_class   = DSC$K_CLASS_S;
	    r_prompt_string.dsc$a_pointer = at_prompt_string;
	    ar_prompt_string = &r_prompt_string;
          }

	  /* argument 2 processing */
	  if (ar_flags == NULL)		/* argument omitted */
	  {
	    al_flags = 0;		/* skip argument */
	  }
	  else
	  {
	    if (ar_flags == Py_None)	/* argument is 'None' */
	    {
	      l_flags  = 0;
	      al_flags = &l_flags;
	    }
	    else
	    {
	      if (!PyInt_Check(ar_flags))
	      {
		PyErr_SetString(PyExc_TypeError,
		    "argument 2: flags must be integer or None");
		return NULL;
	      }
	      else
	      {
		l_flags  = PyInt_AsLong(ar_flags);
		al_flags = &l_flags;
	      }
	    } /* if (ar_flags == Py_None)  -- argument is 'None' */
	  } /* if (ar_flags == NULL) -- argument omitted */
	} /* if (PyArg_ParseTuple(args, "")) */

	/* set up string descriptor for resultant-string argument */
	r_resultant_string.dsc$w_length  = S_RESULTANT_STRING;
	r_resultant_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_resultant_string.dsc$b_class   = DSC$K_CLASS_S;
	r_resultant_string.dsc$a_pointer = &t_resultant_string[0];

	l_status = lib$get_foreign
		(&r_resultant_string	/* resultant-string	*/
		,ar_prompt_string	/* [prompt-string]	*/
		,&w_resultant_length	/* [resultant-length]	*/
		,al_flags		/* [flags]		*/
		);

	if (l_status == SS$_NORMAL)
	{
	    if (al_flags == 0)
	    {
		l_flags = 0;	/* flags are always returned */
	    }
	    /* return resultant-string and flags */
	    return Py_BuildValue
		("s#i", t_resultant_string, (int)w_resultant_length, l_flags);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_get_foreign () */

/* ------------------------------------------------------------------------- */
static char vms_lib_get_symbol__doc[] =
"resultant_string, table_type = vms_lib.get_symbol (symbol)\n\
Get Value of CLI Symbol.";

PyObject *
vms_lib_get_symbol (PyObject *self, PyObject *args)
{
	char			* at_symbol;
	struct dsc$descriptor_s   r_symbol;
	long			  l_symbol_len;

	char			  t_resultant_string[2049];
	struct dsc$descriptor_s   r_resultant_string;
	unsigned short int 	  w_resultant_length;

	char			* at_table_type_indicator;
	long			  l_table_type_indicator;

	long			  l_status;

	if (!PyArg_ParseTuple(args, "s#", &at_symbol, &l_symbol_len))
	{
	    return NULL;
	}

	/* set up descriptor for symbol name */
	r_symbol.dsc$w_length  = l_symbol_len;
	r_symbol.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_symbol.dsc$b_class   = DSC$K_CLASS_S;
	r_symbol.dsc$a_pointer = at_symbol;

	/* set up descriptor for resultant string (output) */
	r_resultant_string.dsc$w_length  = sizeof(t_resultant_string);
	r_resultant_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_resultant_string.dsc$b_class   = DSC$K_CLASS_S;
	r_resultant_string.dsc$a_pointer = &t_resultant_string[0];

	l_status = lib$get_symbol
		(&r_symbol,
		&r_resultant_string,
		&w_resultant_length,
		&l_table_type_indicator
		);

	if (l_status == SS$_NORMAL)
	{
	    /* assign pointer to appropriate table-type text */
	    if (l_table_type_indicator == LIB$K_CLI_LOCAL_SYM)
	    {
		at_table_type_indicator = at_sym_local;
	    }
	    else
	    {
		at_table_type_indicator = at_sym_global;
	    }
	    return Py_BuildValue
		("s#s", t_resultant_string, w_resultant_length
		,at_table_type_indicator);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_get_symbol () */

/* ------------------------------------------------------------------------- */
static char vms_lib_get_ul__doc[] =
"users_language = vms_lib.users_language ()\n\
Return the User's Language.";

PyObject *
vms_lib_get_users_language (PyObject *self, PyObject *args)
{
	char			t_language_string[2049];
	struct dsc$descriptor_s r_language_string;

	long			l_status;

	/* no arguments taken */
	if (!PyArg_ParseTuple(args, ""))
	{
	    return NULL;
	}

	/* set up descriptor for language string */
	r_language_string.dsc$w_length  = sizeof(t_language_string);
	r_language_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_language_string.dsc$b_class   = DSC$K_CLASS_S;
	r_language_string.dsc$a_pointer = &t_language_string[0];

	l_status = lib$get_users_language (&r_language_string);

	if (l_status == SS$_NORMAL)
	{
	    /* remove trailing blanks */
	    {
		long i;
		for (i = (sizeof(t_language_string)-1); i >= 0; i--)
		{
		    if (t_language_string[i] == ' ')
		    {
			t_language_string[i] = '\0';
		    }
		    else
		    {
			break;
		    }
		}
	    }
	    /* return string without trailing <SP> */
	    return Py_BuildValue
		("s#", t_language_string, strlen(t_language_string));
	}
	else
	{
	    if (l_status == LIB$_ENGLUSED)
	    {
		/* use default language: english */
		return Py_BuildValue ("s#", at_english, strlen(at_english));
	    }
	    else
	    {
		PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
		return NULL;
	    }
	}
} /* vms_lib_get_users_language () */

/* ------------------------------------------------------------------------- */
static char vms_lib_lp_lines__doc[] =
"lp_line_count = vms_lib.lp_lines ()\n\
Lines on Each Printer Page.";

PyObject *
vms_lib_lp_lines (PyObject *self, PyObject *args)
{
	long	l_lp_lines;

	/* no arguments taken */
	if (!PyArg_ParseTuple(args, ""))
	{
	    return NULL;
	}

	l_lp_lines = lib$lp_lines ();

	return Py_BuildValue ("i", l_lp_lines);
	/* no error documented */
} /* vms_lib_lp_lines () */

/* ------------------------------------------------------------------------- */
static char vms_lib_put_common__doc[] =
"resultant_length = vms_lib.put_common (source_string)\n\
Put String to Common.";

PyObject *
vms_lib_put_common (PyObject *self, PyObject *args)
{
	char			* at_source_string;
	struct dsc$descriptor_s   r_source_string;
	long			  l_source_string_len;

	unsigned short int 	  w_resultant_length;

	long			  l_status;

	/* request string length because string might contain \0, too */
	if (!PyArg_ParseTuple(args, "s#",
	    &at_source_string, &l_source_string_len))
	{
	    return NULL;
	}

	/* maximum space in the common area is 252 bytes,	 */
	/*   but an error might occur if the high word gets lost */
	if (l_source_string_len > 65535)
	{
	    w_resultant_length = 65535;
	}
	else
	{
	    w_resultant_length = l_source_string_len;
	}

	/* set up descriptor for LIB$PUT_COMMON */
	r_source_string.dsc$w_length  = w_resultant_length;
	r_source_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_source_string.dsc$b_class   = DSC$K_CLASS_S;
	r_source_string.dsc$a_pointer = at_source_string;

	l_status = lib$put_common (&r_source_string, &w_resultant_length);

	/* Warning - string truncation is considered an error, too */
	if (l_status == SS$_NORMAL)
	{
	    return Py_BuildValue ("i", (int)w_resultant_length);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_put_common () */

/* ------------------------------------------------------------------------- */
static char vms_lib_put_output__doc[] =
"vms_lib.put_output (message_string)\n\
Put Line to SYS$OUTPUT.";

PyObject *
vms_lib_put_output (PyObject *self, PyObject *args)
{
	char*			at_message_string;
	struct dsc$descriptor_s r_message_string;
	long			l_message_string_len;

	long			l_status;

	/* request string length because string might contain \0, too */
	if (!PyArg_ParseTuple(args, "s#",
	    &at_message_string, &l_message_string_len))
	{
	    return NULL;
	}

	/* a string descriptor is limited to 65535 characters */
	if (l_message_string_len > 65535)
	{
	    l_message_string_len = 65535;
	}

	/* set up descriptor for LIB$PUT_COMMON */
	r_message_string.dsc$w_length  = l_message_string_len;
	r_message_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_message_string.dsc$b_class   = DSC$K_CLASS_S;
	r_message_string.dsc$a_pointer = at_message_string;

	l_status = lib$put_output (&r_message_string);

	if (l_status == SS$_NORMAL)
	{
	    Py_INCREF(Py_None);
	    return Py_None;
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_put_output () */

/* ------------------------------------------------------------------------- */
static char vms_lib_radix_point__doc[] =
"radix_point_string = vms_lib.radix_point ()\n\
Radix Point Symbol.";

PyObject *
vms_lib_radix_point (PyObject *self, PyObject *args)
{
	char			t_radix_point_string[2049];
	struct dsc$descriptor_s r_radix_point_string;

	unsigned short int	w_resultant_length;

	long			l_status;

	/* no arguments taken */
	if (!PyArg_ParseTuple(args, ""))
	{
	    return NULL;
	}

	/* set up string descriptor to receive radix_point */
	r_radix_point_string.dsc$w_length  = sizeof(t_radix_point_string);
	r_radix_point_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_radix_point_string.dsc$b_class   = DSC$K_CLASS_S;
	r_radix_point_string.dsc$a_pointer = &t_radix_point_string[0];

	l_status = lib$radix_point (&r_radix_point_string, &w_resultant_length);

	if (l_status == SS$_NORMAL)
	{
	    return Py_BuildValue
		("s#", t_radix_point_string, (int)w_resultant_length);
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_radix_point () */

/* ------------------------------------------------------------------------- */
/* @@ code copied from Donn Cave's port */
PyObject *
VMS_LIB_rename_file (PyObject *self, PyObject *args)
{
	char *old, *new;
	dx olddx, newdx;
	int status;

	/* Another stopgap until RMS interface is worked out. */

	if (!PyArg_ParseTuple(args, "ss", &old, &new))
	{
		return NULL;
	}
	initdx(&olddx, old, strlen(old));
	initdx(&newdx, new, strlen(new));
	status = lib$rename_file (&olddx, &newdx);
	if (status & 1) {
		Py_INCREF(Py_None);
		return Py_None;
	} else
	{
		PyVMS_ErrSetVal(vms_lib_Error, 1, status);
		return NULL;
	}
} /* VMS_LIB_rename_file () */

/* ------------------------------------------------------------------------- */
static char vms_lib_reserve_ef__doc[] =
"vms_lib.reserve_ef (event_flag_number)\n\
Reserve Event Flag.";

PyObject *
vms_lib_reserve_ef (PyObject *self, PyObject *args)
{
	long	l_efn;
	long	l_status;

	if (!PyArg_ParseTuple(args, "i", &l_efn))
	{
	    return NULL;
	}

	l_status = lib$reserve_ef (&l_efn);
	if (l_status == SS$_NORMAL)
	{
	    Py_INCREF(Py_None);
	    return Py_None;
	}

	/* error from LIB$RESERVE_EF */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_reserve_ef () */

/* ------------------------------------------------------------------------- */
static char vms_lib_run_program__doc[] =
"vms_lib.run_program (program_name)\n\
Run New Program.";

PyObject *
vms_lib_run_program (PyObject *self, PyObject *args)
{
	char*			at_program_name;
	struct dsc$descriptor_s	r_program_name;
	long			l_program_name_len;

	long			l_status;

	if (!PyArg_ParseTuple(args, "s#",
	     &at_program_name, &l_program_name_len))
	{
	    return NULL;
	}

	/* @@ that does not make much sense because a file specification is */
	/* @@ limited to 255 characters anyway. */
	if (l_program_name_len > 65535)
	{
	    l_program_name_len = 65535;
	}

	/* set up descriptor for LIB$RUN_PROGRAM */
	r_program_name.dsc$w_length  = l_program_name_len;
	r_program_name.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_program_name.dsc$b_class   = DSC$K_CLASS_S;
	r_program_name.dsc$a_pointer = at_program_name;

	l_status = lib$run_program (&r_program_name);

	/* any return from LIB$RUN_PROGRAM indicates an error! */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_run_program () */

/* ------------------------------------------------------------------------- */
struct vmsdef_xr_itmlst3 {
	unsigned short int	w_bufsiz;	/* buffer size		  */
	unsigned short int	w_itmcod;	/* item code		  */
	void			* a_bufadr;	/* buffer address	  */
	void			* aw_retlenaddr;/* returned lenth address */
};
typedef struct vmsdef_xr_itmlst3 vmsdef_xr_itmlst3;

/* reference to translation table for LNM$_name text/code */
extern struct vmsdef_xr_itmtbl VMSDEF_GR_$LNMDEF[];

static char vms_lib_set_logical__doc[] =
"dict = vms_lib.set_logical (logical-name, [value-string], [table] \
[,attributes] [,itemlist])\n\
Set Logical Name.";

/* ---------- */
/* free item-list and any string-buffers from output-items */
static void vms_lib_set_log__free
	(struct vmsdef_xr_itmlst3  * ar_itemlist
	,struct vmsdef_xr_itmtbl  ** ar_itm_vmsdef)
{
	struct vmsdef_xr_itmlst3  * ar_itemlist_ptr;	/* walk along    */
	struct vmsdef_xr_itmtbl  ** ar_itm_vmsdef_ptr;	/* walk along    */

	ar_itemlist_ptr   = ar_itemlist;
	ar_itm_vmsdef_ptr = ar_itm_vmsdef;

	/* loop over item-list */
	while ((ar_itemlist_ptr->w_itmcod != 0) &&
	       (ar_itemlist_ptr->w_bufsiz != 0)   )
	{
	    /* if this is an output-item ... */
	    if ((*ar_itm_vmsdef_ptr)->l_flags & ITMTBL_M_ITMOUT)
	    {
	      /* and it is a string-item */
	      if ((*ar_itm_vmsdef_ptr)->c_typ_lib == 'S')
	      {
		/* drop string buffer, if one is allocated */
		if (ar_itemlist_ptr->a_bufadr != 0)
		{
		  (void) free (ar_itemlist_ptr->a_bufadr);
		}
	      }
	    }
	    ar_itemlist_ptr++;		    /* next item */
	    ar_itm_vmsdef_ptr++;
	}
	/* free the entire VMS item-list */
	(void) free (ar_itemlist);
	return;
} /* vms_lib_set_log__free () */

/* ---------- */
PyObject *
vms_lib_set_logical (PyObject *self, PyObject *args)
{
	char			* at_logical_name;
	struct dsc$descriptor_s   r_logical_name;
	long			  l_logical_name_len;

	char			* at_value_string;
	struct dsc$descriptor_s   r_value_string;
	struct dsc$descriptor_s * ar_value_string;
	long			  l_value_string_len;

	char			* at_table;
	struct dsc$descriptor_s   r_table;
	struct dsc$descriptor_s * ar_table;
	long			  l_table_len;

	PyObject		* ar_py_attributes;
	long			  l_attributes;
	long			* al_attributes;

	/* item-list - tuples of tuples = (('LNM$_name',data), ...) */
	PyObject		* ar_py_item_list;	/* tuples of tuples */

	int			  l_itmlst_size;	/* number of tuples */
	int			  l_itmlst_index;	/* index into "     */
	/* item tuple - ('LNM$_name',data) */
	PyObject		* ar_py_item_tuple;	/* item tuple */

	PyObject		* ar_py_item_code;	/* LNM$_name  */
	char			* at_item_code;		/* "	      */
	unsigned short int	  w_item_code;		/* from VMSDEF */

	/* item 'data' */
	PyObject		* ar_py_item_data;	/* String/L/W/B */

	long			  l_trn_index; /* into VMSDEF_GR_$LNMDEF[] */

	/* VMS item-list to be passed to LIB$SET_LOGICAL() */
	struct vmsdef_xr_itmlst3 * ar_itemlist;		/* start address */
	struct vmsdef_xr_itmlst3 * ar_itemlist_ptr;	/* walk along    */

	/* array that provides data storage for L/W/B integers */
	long			 * al_datalist;
	long			 * al_datalist_ptr;

	/* array that provides storage for returned length for output items */
	unsigned short int	 * aw_retlenlist;
	unsigned short int	 * aw_retlenlist_ptr;

	/* array of pointers from item-list entry into VMSDEF structure */
	/*  this is used for output-item processing */
	struct vmsdef_xr_itmtbl  ** ar_itm_vmsdef;	/* start address */
	struct vmsdef_xr_itmtbl  ** ar_itm_vmsdef_ptr;	/* walk along    */

	long			   l_malloc_bytes; /* for malloc() */

	long			   l_status;
	PyObject*		   ar_dict;	/* dictionary to be returned */
	PyObject*		   ar_dictobj;	/* temp. for insert into " */

	/* -------------------- */
	at_logical_name  = NULL;
	at_value_string  = NULL;
	at_table         = NULL;
	ar_py_attributes = NULL;
	ar_py_item_list  = NULL;
	if (!PyArg_ParseTuple(args, "s#|z#z#OO",
	    &at_logical_name, &l_logical_name_len,
	    &at_value_string, &l_value_string_len,
	    &at_table, &l_table_len,
	    &ar_py_attributes, &ar_py_item_list))
	{
	    return NULL;
	}

	/* argument 1: logical-name */
	/* string length in a VMS descriptor is limited to 65535 chars */
	if (l_logical_name_len > 65535)
	{
	    l_logical_name_len = 65535;
	}
	/* set up string descriptor for LIB$SET_LOGICAL */
	r_logical_name.dsc$w_length  = l_logical_name_len;
	r_logical_name.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_logical_name.dsc$b_class   = DSC$K_CLASS_S;
	r_logical_name.dsc$a_pointer = at_logical_name;

	/* argument 2: value-string (optional/ignored, if item-list) */
	if (at_value_string == NULL)
	{
	    ar_value_string = 0;		/* skip argument */
	}
	else
	{
	    /* string length in a VMS descriptor is limited to 65535 chars */
	    if (l_value_string_len > 65535)
	    {
		l_value_string_len = 65535;
	    }
	    /* set up string descriptor for LIB$SET_LOGICAL */
	    r_value_string.dsc$w_length  = l_value_string_len;
	    r_value_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	    r_value_string.dsc$b_class   = DSC$K_CLASS_S;
	    r_value_string.dsc$a_pointer = at_value_string;
	    ar_value_string = &r_value_string;
	}

	/* argument 3: table (optional) */
	if (at_table == NULL)
	{
	    ar_table = 0;			/* skip argument */
	}
	else
	{
	    /* string length in a VMS descriptor is limited to 65535 chars */
	    if (l_table_len > 65535)
	    {
		l_table_len = 65535;
	    }
	    /* set up string descriptor for LIB$SET_LOGICAL */
	    r_table.dsc$w_length  = l_table_len;
	    r_table.dsc$b_dtype   = DSC$K_DTYPE_T;
	    r_table.dsc$b_class   = DSC$K_CLASS_S;
	    r_table.dsc$a_pointer = at_table;
	    ar_table = &r_table;
	}

	/* argument 4: attributes (optional) */
	if ((ar_py_attributes == NULL)   ||
	    (ar_py_attributes == Py_None)  )
	{
	    al_attributes = 0;			/* skip argument */
	}
	else
	{
	    if (!PyInt_Check(ar_py_attributes))
	    {
		PyErr_SetString(PyExc_TypeError,
		    "argument 4: attributes must be integer or None");
		return NULL;
	    }
	    else
	    {
		l_attributes = PyInt_AsLong(ar_py_attributes);
		al_attributes = &l_attributes;
	    }
	}

	/* -------------------- */
	/* argument 5: item-list (optional) */
	if ((ar_py_item_list == NULL)   ||
	    (ar_py_item_list == Py_None)  )
	{
	    ar_itemlist = 0;		/* skip argument */
	}
	else
	{
	  if (!PyTuple_Check(ar_py_item_list))
	  {
	    PyErr_SetString(PyExc_TypeError,
		"argument 5: item-list must be a tuple of tuples or None");
	    return NULL;
	  }

	  l_itmlst_size = PyObject_Length(ar_py_item_list);

	  /* allocate memory for:
	  ||	- each item-list element
	  ||	- the item-list termination element
	  ||	- an array of longwords for (potentially) each item-list element
	  ||	  unused entries in this array are just ignored
	  ||	- an array of pointers from the item to its corresponding
	  ||	  VMSDEF structure
	  */
	  l_malloc_bytes =
	    ( (l_itmlst_size * sizeof(struct vmsdef_xr_itmlst3)) + /* items */
	      sizeof(struct vmsdef_xr_itmlst3) + /* termination of itemlist */
	      (l_itmlst_size * 4) +		 /* for longword items  % */
	      (l_itmlst_size * 4) +		 /* for VMSDEF pointers % */
	      (l_itmlst_size * 2)		 /* for returned length % */
	    );				/* % - this needs no termination */

	  ar_itemlist = malloc (l_malloc_bytes);
	  if (ar_itemlist == NULL)
	  {
	    PyErr_NoMemory();
	    return NULL;
	  }

	  /* Set the entire allocated memory to 0. */
	  /* This is needed if the loop to build the item-list fails	 */
	  /*  and the cleanup has to free() any output buffers that	 */
	  /*  have been malloc()ed so far. I don't have to set		 */
	  /*  ar_itemlist_ptr->a_bufadr to NULL before building the list */
	  (void) memset (ar_itemlist, 0, l_malloc_bytes);

	  /* ar_itemlist_ptr is used to walk along the individual elements */
	  /*  in the VMS item-list */
	  ar_itemlist_ptr = ar_itemlist;	/* used to walk over list */

	  /* this is an array of longword entries than can provide storage */
	  /*  for Longword, Word and Byte data buffers */
	  /* al_datalist_ptr is used to walk along the individual elements */
	  /*  in the item-list and calculate the address of the next */
	  /*  integer data buffer */
	  /* if a string-buffer is used, then the entry is skipped */
	  al_datalist = (long *) (ar_itemlist + (l_itmlst_size+1));
						/* start of array  */
					/* '+1' for itemlist termination */
	  al_datalist_ptr = al_datalist;	/* used to walk over array */

	  /* this is an array of pointers from an item-list entry into the */
	  /*  VMSDEF structure */
	  ar_itm_vmsdef = (void *)(al_datalist_ptr + l_itmlst_size);
						/* start of array  */
	  ar_itm_vmsdef_ptr = ar_itm_vmsdef;	/* used to walk over array */

	  /* this is an array of word entries than can provide storage */
	  /*  for returned lengths for output items */
	  /* aw_retlenlist_ptr is used to walk along the individual elements */
	  /*  in the item-list and calculate the address of the next */
	  /*  returned length data buffer */
	  /* for input-items the entry is skipped */
	  aw_retlenlist = (unsigned short *) (ar_itm_vmsdef + l_itmlst_size);
						/* start of array  */
	  aw_retlenlist_ptr = aw_retlenlist;	/* used to walk over array */

	  /* -------------------- */
	  /* loop over each item-list tuple */
	  for (l_itmlst_index = 0; l_itmlst_index < l_itmlst_size; l_itmlst_index++)
	  {
	    /* get item-list element */
	    ar_py_item_tuple = PyTuple_GetItem(ar_py_item_list, l_itmlst_index);
	    if (ar_py_item_tuple == NULL)
	    {
		/* free item-list *and* all malloc()ed string-buffers */
		(void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
		return NULL;
	    }
	    /* each list element must be a tuple itself ('LNM$_name,data) */
	    if (!PyTuple_Check(ar_py_item_tuple))
	    {
		/* free item-list *and* all malloc()ed string-buffers */
		(void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
		PyErr_SetString(PyExc_TypeError,
		    "argument 5: tuple-element:n is not a tuple");
		return NULL;
	    }
	    /* -------------------- */
	    /* get item-code from tuple */
	    ar_py_item_code = PyTuple_GetItem(ar_py_item_tuple, 0);
	    if (PyErr_Occurred())
	    {
		/* free item-list *and* all malloc()ed string-buffers */
		(void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
		return NULL;
	    }
	    /* item-code must be a string ('LNM$_name') */
	    if (!PyString_Check(ar_py_item_code))
	    {
		/* free item-list *and* all malloc()ed string-buffers */
		(void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
		PyErr_SetString(PyExc_TypeError,
		    "argument 5: tuple-element:n item-code is not a string");
		return NULL;
	    }

	    /* -------------------- */
	    /* convert item-code string to LNM$_ numerical value */
	    l_trn_index  = 0;
	    at_item_code = PyString_AsString(ar_py_item_code);
	    while (VMSDEF_GR_$LNMDEF[l_trn_index].w_vmsvermin != 0)
	    {
	      /* check VMS version */
	      if ((VMSDEF_GR_$LNMDEF[l_trn_index].w_vmsvermin <= vmsdef_gw_vmsver) &&
		  (VMSDEF_GR_$LNMDEF[l_trn_index].w_vmsvermax >= vmsdef_gw_vmsver)  )
	      {
		/* compare item name */
		if (strcmp(at_item_code
			  ,VMSDEF_GR_$LNMDEF[l_trn_index].at_itmnam_py
			  ) == 0)
		{
		  /* found textual match - use numerical value of item-code */
		  w_item_code = VMSDEF_GR_$LNMDEF[l_trn_index].w_itmcod;
		  break;				/* leave loop */
		} /* item code string match */
	      } /* w_vmsvermin / w_vmsvermax match with vmsdef_gw_vmsver */
	      l_trn_index++;				/* check next item code */
	    } /* while (VMSDEF_GR_$LNMDEF[l_trn_index].w_vmsvermin != 0) */

	    /* ---------------------------------------- */
	    if (VMSDEF_GR_$LNMDEF[l_trn_index].w_vmsvermin == 0)
	    {
	      /* item code not found - abort */
	      /* free item-list *and* all malloc()ed string-buffers */
	      (void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
	      PyErr_SetString(PyExc_ValueError, "argument 5: unknown LNM$_ item code");
	      return NULL;
	    }

	    /* ---------------------------------------- */
/* printf ("@@itm:%s=%d\n", PyString_AsString(ar_py_item_code),
		(int)w_item_code); */

	    /* store numerical item-code into VMS item-list */
	    ar_itemlist_ptr->w_itmcod = w_item_code;

	    /* ---------------------------------------- */
	    /* if item-code is an input-item */
	    if (VMSDEF_GR_$LNMDEF[l_trn_index].l_flags & ITMTBL_M_ITMINP)
	    {
	      /* get data portion of the item tuple */
	      ar_py_item_data = PyTuple_GetItem(ar_py_item_tuple, 1);
	      if (PyErr_Occurred())
	      {
		/* free item-list *and* all malloc()ed string-buffers */
		(void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
		return NULL;
	      }

	      /* data type */
	      if (VMSDEF_GR_$LNMDEF[l_trn_index].c_typ_lib == 'S')
	      {
		/* check data-type with LNMDEF */
		if (!PyString_Check(ar_py_item_data))
		{
		  /* free item-list *and* all malloc()ed string-buffers */
		  (void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
		  PyErr_SetString(PyExc_TypeError,
		    "argument 5: tuple-element:n item-buffer is not a string");
		  return NULL;
		}
		/* input item: take length from object */
		ar_itemlist_ptr->w_bufsiz = PyObject_Length(ar_py_item_data);
		/* string object: just point to storage in the object */
		/* @@@ Warning: this is not thread-safe */
		ar_itemlist_ptr->a_bufadr = PyString_AsString(ar_py_item_data);
	      }
	      else
	      {
		/* assume the data-type of the buffer is L, W or B */
		if (!PyInt_Check(ar_py_item_data))
		{
		  /* free item-list *and* all malloc()ed string-buffers */
		  (void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
		  PyErr_SetString(PyExc_TypeError,
		    "argument 5: tuple-element:n item-buffer is not an integer");
		  return NULL;
		}
		/* -------------------- */
		/* use the size from the VMSDEF table */
		ar_itemlist_ptr->w_bufsiz = VMSDEF_GR_$LNMDEF[l_trn_index].w_bufsiz;

		/* address of binary data (in longword array) */
		ar_itemlist_ptr->a_bufadr = al_datalist_ptr;

		/* store data in array of longwords */
		*al_datalist_ptr = PyInt_AsLong(ar_py_item_data);
	      }
	      ar_itemlist_ptr->aw_retlenaddr = 0;	/* ignore */
	      /* ---------------------------------------- */
	    }
	    else
	    {
	      /* ASSUME this is an output-element, provide a buffer */
	      /* ---------------------------------------- */
	      /* data type */
	      if (VMSDEF_GR_$LNMDEF[l_trn_index].c_typ_lib == 'S')
	      {
		/* allocate storage for a string */
		ar_itemlist_ptr->a_bufadr =
		  malloc ((int)VMSDEF_GR_$LNMDEF[l_trn_index].w_bufsiz);
		if (ar_itemlist_ptr->a_bufadr == NULL)
		{
		  /* free item-list *and* all malloc()ed string-buffers */
		  (void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);

		  PyErr_NoMemory();
		  return NULL;
		}
	      }
	      else
	      {
		/* address of binary data (in longword array) */
		ar_itemlist_ptr->a_bufadr = al_datalist_ptr;
	      }
	      /* set buffer size */
	      ar_itemlist_ptr->w_bufsiz = VMSDEF_GR_$LNMDEF[l_trn_index].w_bufsiz;
	      /* provide storage for returned length */
	      ar_itemlist_ptr->aw_retlenaddr = aw_retlenlist_ptr;
	      /* ---------------------------------------- */
	    } /*if (VMSDEF_GR_$LNMDEF[l_trn_index].l_flags & ITMTBL_M_ITMINP)*/

	    /* always store pointer to VMSDEF entry for this item-code */
	    /* because this is used in a decision to build the dictionary */
	    /* which is returned to Python */
	    *ar_itm_vmsdef_ptr = &VMSDEF_GR_$LNMDEF[l_trn_index];

	    /* ---------------------------------------- */
	    ar_itemlist_ptr++;		/* next address of itemlist element */
	    al_datalist_ptr++;		/* next address of item-buffer */
	    ar_itm_vmsdef_ptr++;	/* next pointer into VMSDEF */
	    aw_retlenlist_ptr++;	/* next addr. of returned length buf.*/
	  } /* for (l_itmlst_index = 0; l_itmlst_index < l_itmlst_size; .. */

	  /* terminate VMS item-list */
	  ar_itemlist_ptr->w_itmcod = 0;
	  ar_itemlist_ptr->w_bufsiz = 0;
	  /* Note: the other pointers don't need termination	     */
	  /* Warning: there is no storage allocated for termination! */
	} /* if ((ar_py_item_list == NULL||Py_None)) */

	/* ---------------------------------------- */
	l_status = lib$set_logical
		(&r_logical_name
		,ar_value_string
		,ar_table
		,al_attributes
		,ar_itemlist
		);

	/* -------------------- */
	/* create a dictionary because the user might have specified an */
	/*  output-item in the item-list */
	ar_dict = PyDict_New();
	if (ar_dict == NULL)
	{
	    /* free item-list *and* all malloc()ed string-buffers */
            (void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
	    return NULL;
	}

	/* always return the status code - SS$_SUPERSEEDE is also success */
	ar_dictobj = PyInt_FromLong(l_status);
	if (ar_dictobj == NULL)
	{
	    Py_DECREF(ar_dict);		/* drop dictionary */
	    /* free item-list *and* all malloc()ed string-buffers */
            (void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
	    return NULL;
	}
	PyDict_SetItemString(ar_dict, "status", ar_dictobj);
	Py_DECREF(ar_dictobj);

	/* build remaining items only if 'success' */
	if ((l_status == SS$_NORMAL) || (l_status == SS$_BUFFEROVF) ||
	    (l_status == SS$_SUPERSEDE)					)
	{
	  /* loop over item-list (if specified) and return any output items */
	  if (ar_itemlist != 0)
	  {
	    /* re-position pointers to begin of each list */
	    ar_itemlist_ptr   = ar_itemlist;	/* VMS item-list	*/
	    al_datalist_ptr   = al_datalist;	/* L/W/B integer data	*/
	    ar_itm_vmsdef_ptr = ar_itm_vmsdef;	/* ptr to item's VMSDEF */
	    aw_retlenlist_ptr = aw_retlenlist;	/* ptr to returned length */

	    /* loop over item-list */
	    while ((ar_itemlist_ptr->w_itmcod != 0) &&
		   (ar_itemlist_ptr->w_bufsiz != 0)   )
	    {
	      /* if this is an output-item ... */
	      if ((*ar_itm_vmsdef_ptr)->l_flags & ITMTBL_M_ITMOUT)
	      {
		/* put it into the dictionary */
		if ((*ar_itm_vmsdef_ptr)->c_typ_lib == 'S')
		{
		  /* it's a string */
		  ar_dictobj = Py_BuildValue ("s#",
			ar_itemlist_ptr->a_bufadr,	/* data */
			(int)*aw_retlenlist_ptr);	/* size */
		}
		else
		{
		  /* assume it's an integer */
		  ar_dictobj = Py_BuildValue ("i",
			(int)*al_datalist_ptr);
		  /* size = *aw_retlenlist_ptr */
		}
		if (ar_dictobj == NULL)
		{
		  Py_DECREF(ar_dict);		/* drop dictionary */
		  /* free item-list *and* all malloc()ed string-buffers */
		  (void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
		  return NULL;
		}

		/* store item-name ('LNM$_name') + data into dictionary */
		PyDict_SetItemString(ar_dict, (*ar_itm_vmsdef_ptr)->at_itmnam_py,
		    ar_dictobj);
		Py_DECREF(ar_dictobj);
	      }
	      ar_itemlist_ptr++;		    /* next item */
	      al_datalist_ptr++;
	      ar_itm_vmsdef_ptr++;
	      aw_retlenlist_ptr++;
	    }
	    /* free item-list *and* any string-buffers from output-items  */
            (void) vms_lib_set_log__free(ar_itemlist, ar_itm_vmsdef);
	  } /* if (ar_itemlist != 0) */
	} /* if (l_status == (SS$_NORMAL || SS$_BUFFEROVF || SS$_SUPERSEDE)) */

	/* done, return to Python with the dictionary */
	return ar_dict;
} /* vms_lib_set_logical () */

/* ------------------------------------------------------------------------- */
static char vms_lib_set_symbol__doc[] =
"vms_lib.set_symbol (symbol, value-string [,table-type-indicator])\n\
Set Value of CLI Symbol.";

PyObject *
vms_lib_set_symbol (PyObject *self, PyObject *args)
{
	char			* at_symbol;
	struct dsc$descriptor_s   r_symbol;
	long			  l_symbol_len;

	char			* at_value_string;
	struct dsc$descriptor_s   r_value_string;
	long			  l_value_string_len;

	char			* at_table_type_indicator;
	long			  l_table_type_indicator;

	long			  l_status;

	/* provide default value for table-type-indicator, which is optional */
	at_table_type_indicator = at_sym_local;

	/* allow string or None ----------v */
	if (!PyArg_ParseTuple(args, "s#s#|z",
	    &at_symbol, &l_symbol_len,
	    &at_value_string, &l_value_string_len,
	    &at_table_type_indicator
	    ))
	{
	    return NULL;
	}

	if (at_table_type_indicator == NULL)	/* 'None' given */
	{
	    at_table_type_indicator = at_sym_local;
	}

	/* that might not make much sense because a symbol name is */
	/*  limited to 255 characters. */
	if (l_symbol_len > 65535)
	{
	    l_symbol_len = 65535;
	}

	/* set up descriptor for symbol name */
	r_symbol.dsc$w_length  = l_symbol_len;
	r_symbol.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_symbol.dsc$b_class   = DSC$K_CLASS_S;
	r_symbol.dsc$a_pointer = at_symbol;

	/* that might not make much sense because a symbol name is */
	/*  limited to 255 characters. */
	if (l_value_string_len > 65535)
	{
	    l_value_string_len = 65535;
	}

	/* set up descriptor for equivalence (value) string */
	r_value_string.dsc$w_length  = l_value_string_len;
	r_value_string.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_value_string.dsc$b_class   = DSC$K_CLASS_S;
	r_value_string.dsc$a_pointer = at_value_string;

	/* check table indicator */
	if (strcmp(at_table_type_indicator, at_sym_global) == 0)
	{
	    l_table_type_indicator = LIB$K_CLI_GLOBAL_SYM;
	}
	else
	{
	    if (strcmp(at_table_type_indicator, at_sym_local) == 0)
	    {
		l_table_type_indicator = LIB$K_CLI_LOCAL_SYM;
	    }
	    else
	    {
		l_status = LIB$_INVARG;
		PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
		return NULL;
	    }
	}

	l_status = lib$set_symbol
			(&r_symbol, &r_value_string, &l_table_type_indicator);
	if (l_status == SS$_NORMAL)
	{
	    Py_INCREF(Py_None);
	    return Py_None;
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_set_symbol () */

/* ------------------------------------------------------------------------- */
/* @@ code copied from Donn Cave's port */
PyObject *
VMS_LIB_spawn (PyObject *self, PyObject *args)
{
	/*
	 * command
	 * input
	 * output
	 * flags
	 * name
	 */
	char *cmdp, *inputp, *outputp, *namep;
	int flags;
	dx cmddx, inpdx, outdx, namdx;
	dx *cmd, *inp, *out, *nam;
	int status;

	cmd = inp = out = nam = 0;
	cmdp = inputp = outputp = namep = 0;
	flags = 0;
	if (args) {
	    if (PyString_Check(args))
		cmdp = PyString_AsString(args);
	    else if (!PyArg_ParseTuple(args, "s|ssls",
			&cmdp, &inputp, &outputp, &flags, &namep))
		return NULL;
	    if (cmdp) {
		cmd = &cmddx;
		initdx(cmd, cmdp, strlen(cmdp));
	    }
	    if (inputp) {
		inp = &inpdx;
		initdx(inp, inputp, strlen(inputp));
	    }
	    if (outputp) {
		out = &outdx;
		initdx(out, outputp, strlen(outputp));
	    }
	    if (namep) {
		nam = &namdx;
		initdx(nam, namep, strlen(namep));
	    }
	}

	/*
	 * Might think about returning an object that can be used to
	 * check status, for NOWAIT option.
	 */

	status = lib$spawn(cmd, inp, out, flags, nam);
	if (status & 1) {
		Py_INCREF(Py_None);
		return Py_None;
	}
	else
	{
		PyVMS_ErrSetVal(vms_lib_Error, 1, status);
		return NULL;
	}
} /* VMS_LIB_spawn () */

/* ------------------------------------------------------------------------- */
static char vms_lib_sub_times__doc[] =
"resultant_time = vms_lib.sub_times (time1, time2)\n\
Subtract Two Quadword Times.";

PyObject *
vms_lib_sub_times (PyObject *self, PyObject *args)
{
	PyObject	* ar_time1;
	long		  q_time1[2];

	PyObject	* ar_time2;
	long		  q_time2[2];

	long		  q_resultant_time[2];

	long		  l_status;

	if (!PyArg_ParseTuple(args, "OO", &ar_time1, &ar_time2))
	{
	    return NULL;
	}

	/* ---------- */
	if (!PyLong_Check(ar_time1))
	{
	    PyErr_SetString(PyExc_ValueError, "argument 1: must be long int");
	    return NULL;
	}

	/* convert PyLongInt to binary quadword */
	l_status = vms__cvt_pylong2quad (ar_time1, &q_time1[0]);
	if (l_status == -1)
	{
	    /* error string has been set */
	    return NULL;
	}
	
	/* ---------- */
	if (!PyLong_Check(ar_time2))
	{
	    PyErr_SetString(PyExc_ValueError, "argument 2: must be long int");
	    return NULL;
	}

	/* convert Py long int to binary quadword */
	l_status = vms__cvt_pylong2quad (ar_time2, &q_time2[0]);
	if (l_status == -1)
	{
	    /* error string has been set */
	    return NULL;
	}

	/* ---------- */
	l_status = lib$sub_times
	    (&q_time1[0], &q_time2[0], &q_resultant_time[0]);
	if (l_status == LIB$_NORMAL)	/* note: _not_ SS$_NORMAL !! */
	{
	    /* convert binary quadword to py long int */
	    return vms__cvt_quad2pylong (&q_resultant_time[0]);
	}

	/* ---------- */

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_sub_times () */

/* ------------------------------------------------------------------------- */
static char vms_lib_wait__doc[] =
"vms_lib.wait (seconds)\n\
Wait a Specified Period of Time.";

PyObject *
vms_lib_wait (PyObject *self, PyObject *args)
{
	float	f_wait_time;
	long	l_status;

	if (!PyArg_ParseTuple(args, "f", &f_wait_time))
	{
	    return NULL;
	}

	l_status = lib$wait (&f_wait_time);

	if (l_status == SS$_NORMAL)
	{
	    Py_INCREF(Py_None);
	    return Py_None;
	}

	/* failed - error */
	PyVMS_ErrSetVal(vms_lib_Error, 1, l_status);
	return NULL;
} /* vms_lib_wait () */

/* ------------------------------------------------------------------------- */
static struct PyMethodDef vms_lib_methods[] = {
{"add_times",	       (PyCFunction)vms_lib_add_times,		METH_VARARGS, vms_lib_add_times__doc},
{"asn_wth_mbx",        (PyCFunction)vms_lib_asn_wth_mbx,	METH_VARARGS, vms_lib_asn_wth_mbx__doc},
{"attach",             (PyCFunction)vms_lib_attach,             METH_VARARGS, vms_lib_attach__doc},
{"create_dir",         (PyCFunction)vms_lib_create_dir,         METH_VARARGS, vms_lib_create_dir__doc},
{"currency",           (PyCFunction)vms_lib_currency,           METH_VARARGS, vms_lib_currency__doc},
{"cvt_vectim",         (PyCFunction)vms_lib_cvt_vectim,         METH_VARARGS, vms_lib_cvt_vectim__doc},
{"date_time",          (PyCFunction)vms_lib_date_time,          METH_VARARGS, vms_lib_date_time__doc},
{"day",                (PyCFunction)vms_lib_day,                METH_VARARGS, vms_lib_day__doc},
{"day_of_week",        (PyCFunction)vms_lib_day_of_week,        METH_VARARGS, vms_lib_day_of_week__doc},
{"delete_file",        (PyCFunction)VMS_LIB_delete_file,        METH_VARARGS},
{"delete_logical",     (PyCFunction)vms_lib_delete_logical,     METH_VARARGS, vms_lib_delete_logical__doc},
{"delete_symbol",      (PyCFunction)vms_lib_delete_symbol,      METH_VARARGS, vms_lib_delete_symbol__doc},
{"digit_sep",          (PyCFunction)vms_lib_digit_sep,          METH_VARARGS, vms_lib_digit_sep__doc},
{"do_command",	       (PyCFunction)vms_lib_do_command,         METH_VARARGS, vms_lib_do_command__doc},
{"fid_to_name",        (PyCFunction)vms_lib_fid_to_name,	METH_VARARGS, vms_lib_fid_to_name__doc},
{"find_image_symbol",  (PyCFunction)vms_lib_find_image_symbol,  METH_VARARGS, vms_lib_find_image_symbol__doc},
{"free_ef",            (PyCFunction)vms_lib_free_ef,            METH_VARARGS, vms_lib_free_ef__doc},
{"getdvi",             (PyCFunction)vms_lib_getdvi,             METH_VARARGS, vms_lib_getdvi__doc},
{"getjpi",             (PyCFunction)vms_lib_getjpi,             METH_VARARGS, vms_lib_getjpi__doc},
/* getqui: 06000 */
{"getqui",             (PyCFunction)vms_lib_getqui,             METH_VARARGS, vms_lib_getqui__doc},
{"getsyi",             (PyCFunction)vms_lib_getsyi,             METH_VARARGS, vms_lib_getsyi__doc},
{"get_command",        (PyCFunction)vms_lib_get_command,        METH_VARARGS, vms_lib_get_command__doc},
{"get_common",         (PyCFunction)vms_lib_get_common,         METH_VARARGS, vms_lib_get_common__doc},
{"get_ef",             (PyCFunction)vms_lib_get_ef,             METH_VARARGS, vms_lib_get_ef__doc},
{"get_foreign",        (PyCFunction)vms_lib_get_foreign,        METH_VARARGS, vms_lib_get_foreign__doc},
{"get_symbol",         (PyCFunction)vms_lib_get_symbol,         METH_VARARGS, vms_lib_get_symbol__doc},
{"get_users_language", (PyCFunction)vms_lib_get_users_language, METH_VARARGS, vms_lib_get_ul__doc},
{"lp_lines",           (PyCFunction)vms_lib_lp_lines,           METH_VARARGS, vms_lib_lp_lines__doc},
{"put_common",         (PyCFunction)vms_lib_put_common,         METH_VARARGS, vms_lib_put_common__doc},
{"put_output",         (PyCFunction)vms_lib_put_output,         METH_VARARGS, vms_lib_put_output__doc},
{"radix_point",        (PyCFunction)vms_lib_radix_point,        METH_VARARGS, vms_lib_radix_point__doc},
/* @@ rename_file */
{"reserve_ef",         (PyCFunction)vms_lib_reserve_ef,         METH_VARARGS, vms_lib_reserve_ef__doc},
{"run_program",        (PyCFunction)vms_lib_run_program,        METH_VARARGS, vms_lib_run_program__doc},
{"set_logical",        (PyCFunction)vms_lib_set_logical,        METH_VARARGS, vms_lib_set_logical__doc},
{"set_symbol",         (PyCFunction)vms_lib_set_symbol,         METH_VARARGS, vms_lib_set_symbol__doc},
{"spawn",              (PyCFunction)VMS_LIB_spawn,              METH_VARARGS},
{"sub_times",          (PyCFunction)vms_lib_sub_times,          METH_VARARGS, vms_lib_sub_times__doc},
{"wait",               (PyCFunction)vms_lib_wait,               METH_VARARGS, vms_lib_wait__doc},
{NULL,                 NULL}
};

void
initvms_lib()
{
	PyObject *m, *d;

	m = Py_InitModule4("vms_lib",
			   vms_lib_methods,
			   vms_lib_doc__,
			   (PyObject *)NULL,
			   PYTHON_API_VERSION);
	d = PyModule_GetDict(m);

	/* ----- ----- ----- ----- */
	/* Initialize exception */
	vms_lib_Error = PyErr_NewException("vms_lib.error", NULL, NULL);
	if (vms_lib_Error == NULL || PyDict_SetItemString(d, "error", vms_lib_Error) != 0)
	    Py_FatalError("can't define vms_lib.error");

	/* ----- ----- ----- ----- */
	/* Check for errors */
        if (PyErr_Occurred())
	{
	    Py_FatalError("can't initialize module vms_lib");
	}

	/* ----- ----- ----- ----- */
} /* initvms_lib () */

/* ------------------------------------------------------------------------- */

/* EOF: VMS_LIB.C */
