#ifndef _VMSDEF_H_
#define _VMSDEF_H_

/* VMSDEF.H -- 30-SEP-1996 Uwe Zessin
   06-SEP-1997 ZE. -- add new item-code table format
   30-SEP-1997 ZE. -- rename from VMS_DEFTABLE.H to VMSDEF.H
   16-OCT-1997 ZE. -- Rename field 'w_vmsver' to 'w_vmsvermin' and add field
			'w_vmsvermax' because some data can change from
			version to version.
   17-OCT-1997 ZE. -- add data structure for directory of VMSDEF tables
   29-NOV-1997 ZE. -- changes for Python V1.5B1
   03-JAN-1998 ZE. -- additional data structure for constants/bitmasks that
			don't map to item codes
   11-JAN-1998 ZE. -- change/add ITMTBL_M_*
   25-APR-1998 ZE. -- add PYVMS_XR_IDSC 'ImageDeSCriptor'
		(has nothing to do with 'classical' VMSDEF but I don't want to
		create many include files)
*/

/* ---------------------------------------- */

#ifdef __cplusplus
    extern "C" {
#endif

#include "python.h"

#if defined(__DECC) || defined(__DECCXX)
#pragma __member_alignment __save
#pragma __member_alignment
#endif

/* ---------------------------------------- */

void PyVMS_ErrSetVal(PyObject *error, int n, int cond, ...);

/* ---------------------------------------- */

#define K_ITM_NONE    0	/* not an item, but in the same structure */
#define K_ITM_BOOLEAN 1
#define K_ITM_INTEGER 2
#define K_ITM_LONG    2
#define K_ITM_STRING  3

/* ---------------------------------------- */
typedef unsigned short int vmsdef_xw_vmsver;

/* ---------------------------------------- */
/* new format of external MACRO module definition */

struct vmsdef_xr_itmtbl {
	unsigned short int	w_itmcod;	/* VMS item code */
	vmsdef_xw_vmsver	w_vmsvermin;	/* minimum VMS version */
	vmsdef_xw_vmsver	w_vmsvermax;	/* maximum VMS version */
	unsigned short int	w_filler;	/*   keep alignment   */
	char*			at_itmnam_py;	/* Python item name */
	void*			ar_bitmsk;	/* pointer to vmsdef_x_bitmsk */
	unsigned int		l_flags;	/* see below */
	unsigned short int	w_bufsiz;	/* item buffer size */
	char			c_typ_lib;	/* data type LIB$name() */
	char			c_typ_sys;	/* data type SYS$name() */
};
typedef struct vmsdef_xr_itmtbl vmsdef_xr_itmtbl;

/* contents of L_FLAGS -- synchronize names + values with VMSDEF2MAR.COM and
   VMSVER2MAR.COM ! */
#define ITMTBL_M_ARCH_VAX    1   /* item valid on VAX architecture   */
#define ITMTBL_M_ARCH_ALPHA  2   /* item valid on Alpha architecture */
#define ITMTBL_M_GETSYI_SYSPAR 4 /* $GETSYI() system parameter */
#define ITMTBL_M_GETSYI_SYSVAL 8 /* $GETSYI() system value     */
#define ITMTBL_M_GETQUI_FNCREQ 4 /* $GETQUI() - request function code */
#define ITMTBL_M_ITMOUT  8	 /* output item code */
#define ITMTBL_M_ITMINP 16	 /* input item code */
/* #define ITMTBL_M_MSK64BIT 32 ? 64-bit mask ? */

/* ---------------------------------------- */
/* define bitmask element - pointed to by VMSDEF_XR_ITMTBL.AR_BITMSK */
struct vmsdef_x_bitmsk {
	unsigned int		l_bit_value;	/* bit value (1,2,4,8,...) */
	char*			at_bit_name;	/* bit name */
	vmsdef_xw_vmsver	w_vmsvermin;	/* minimum VMS version */
	vmsdef_xw_vmsver	w_vmsvermax;	/* maximum VMS version */
};
typedef struct vmsdef_x_bitmsk vmsdef_x_bitmsk;

/* ---------------------------------------- */
/* translation table for VMS version strings to integer version number */
/* (no maximum/minimum version number) */

struct vmsdef_xr_vmsvertbl {
	vmsdef_xw_vmsver	w_vmsver;	/* VMS version number */
	unsigned short int	w_filler;
	char*			at_vmsver;	/* VMS version string */
	unsigned int		l_flags;	/* see above ITMTBL_M_ARCH_* */
};
typedef struct vmsdef_xr_vmsvertbl vmsdef_xr_vmsvertbl;

/* ---------------------------------------- */
/* contents of L_TBLTYP -- synchronize names + values with VMSDEF2MAR.COM and
   VMSVER2MAR.COM ! */
#define CONMSKTBL_K_BITMASK32  1
#define CONMSKTBL_K_CONSTANT32 2

/* data structure for constants and bitmask tables pointer */
struct vmsdef_xr_conmsktbl {
	void*			ar_bitmsk;	/* pointer to vmsdef_x_bitmsk */
	long			l_tbltyp;	/* CONMSKTBL_K_xxx */
};
typedef struct vmsdef_xr_conmsktbl vmsdef_xr_conmsktbl;

/* ---------------------------------------- */
/* data structure for directory of VMSDEF tables */

struct vmsdef_xr_defdirtbl {
	void*			ar_itmtbl;	/* addr of table */
	void*			ar_conmsktbl;	/* constants and masks */
	char*			at_vmsdefnam;	/* xxxDEF name */
};
typedef struct vmsdef_xr_defdirtbl vmsdef_xr_defdirtbl;

/* ---------------------------------------- */

/* PYVMS_XR_IDSC 'ImageDeSCriptor' */

struct vmsdef_xr_idsc {
	void*			a_next;		/* next IDSC */
	vmsdef_xw_vmsver	w_vmsver;	/* VMS version */
	short int		w_fill;
	void*			ar_vmslbr;	/* vms_lbr functions */
	void*			ar_vmslib;	/* vms_lib functions */
	void*			ar_vmssys;	/* vms_sys functions */
				/* layout not worked out */
};
typedef struct vmsdef_xr_idsc vmsdef_xr_idsc;

/* ---------------------------------------- */

#ifdef __cplusplus
    }
#endif

#if defined(__DECC) || defined(__DECCXX)
#pragma __member_alignment __restore
#endif

/* ---------------------------------------- */

#endif /* _VMSDEF_H_ */

/* EOF: VMSDEF.H */
